"""
How an external solver (eg: Clingo, but could be Z3 etc) integrates into Instal
"""
##-- imports
from __future__ import annotations

import abc
import logging as logmod
import time
import warnings
from collections import defaultdict
from dataclasses import InitVar, dataclass, field
from typing import (IO, TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, List, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)

##-- end imports

logging = logmod.getLogger(__name__)

from clingo import parse_term
from jgdv.apis.clingo.ast import SolverAST

@dataclass
class SolverModelResult:
    """
    The immediate results data structure returned by a solver.
    Does no translation from the data structures the solver uses.

    ie: for Clingo, it is lists of clingo.Symbol's
    """
    atoms   : list[Any]
    shown   : list[Any]
    cost    : float
    number  : int
    optimal : bool
    type    : Any

@dataclass
class SolverWrapper_i:
    """
    An wrapper around a solver (ie: clingo) to interface with the rest of instal
    """

    program        : None|str                = field(default=None)
    input_files    : list[Path]              = field(default_factory=list, kw_only=True)

    timestamp      : float                   = field(init=False, default_factory=time.time)
    results        : list[SolverModelResult] = field(init=False, default_factory=list)
    current_answer : int                     = field(init=False, default=0)
    cycle          : int                     = field(init=False, default=0)
    observations   : list[SolverAST]           = field(default_factory=list)

    def __post_init__(self): pass

    @abc.abstractmethod
    def solve(self, assertions:None|list[Any]=None, fresh=False) -> int: pass

    @property
    @abc.abstractmethod
    def metadata(self): pass



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import importlib
from tomlguard import TomlGuard

PAD           : Final[int] = 15

@dataclass
class JGDVTraceRecord:
    message : str                      = field()
    flags   : None|ReportEnum          = field()
    args    : list[Any]                = field(default_factory=list)
    time    : datetime.datetime        = field(default_factory=datetime.datetime.now)

    def __str__(self):
        match self.message:
            case str():
                return self.message.format(*self.args)
            case DootTaskSpec():
                return str(self.message.name)
            case _:
                return str(self.message)

    def __contains__(self, other:ReportEnum) -> bool:
        return all([x in self.flags for x in other])

    def __eq__(self, other:ReportEnum) -> bool:
        return self.flags == other

    def some(self, other:reportPositionEnum) -> bool:
        return any([x in self.flags for x in other])



#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

from jgdv.files_formats.tags.base import TagFile

@dataclass
class IndexFile(TagFile):
    """ Utility class for index file specification and writing """

    mapping : Dict[str, Set[pl.Path]] = field(default_factory=lambda: defaultdict(set))
    ext     : str                 = field(default=".index")

    def __iadd__(self, value) -> IndexFile:
        return self.update(value)

    def __str__(self):
        """
        Export the mapping, 1 key per line, as:
        `key` : `len(values)` : ":".join(`values`)
        """
        key_sort = sorted(list(self.mapping.keys()))
        total = [self.sep.join([k, str(self.counts[k])]
                               + sorted(str(y) for y in self.mapping[k]))
                 for k in key_sort]
        return "\n".join(total)

    def __repr__(self):
        return f"<{self.__class__.__name__}: {len(self)}>"

    def update(self, *values):
        for val in values:
            match val:
                case (str() as key, maybecount, *rest):
                    paths = set(pl.Path(x) for x in rest)
                    try:
                        count = int(maybecount)
                    except ValueError:
                        paths.add(pl.Path(count))
                        count = len(paths)

                    norm_key = self._inc(key, amnt=count)
                    self.mapping[norm_key].update(paths)
                case _:
                    raise TypeError("Unexpected form in index update", val)

        return self

    def files_for(self, *values, op="union"):
        the_files = None
        match op:
            case "union":
                fn = lambda x, y: x & y
            case "diff":
                fn = lambda x, y: x | y
            case "xor":
                fn = lambda x, y: x ^ y
            case "rem":
                fn = lambda x, y: x - y
            case _:
                raise TypeError("Bad Op specified: ", op)

        for val in values:
            match the_files:
                case None if val in self.mapping:
                    the_files = self.mapping[val]
                case None:
                    continue
                case {} if op != "union":
                    return the_files
                case _:
                    the_files = fn(the_files, self.mapping[val])

        return the_files



#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

from jgdv.files_formats.tags.base import TagFile

@dataclass
class SubstitutionFile(TagFile):
    """ SubstitutionFiles add a replacement tag for some tags """

    ext           : str                  = field(default=".sub")
    substitutions : Dict[str, set[str]] = field(default_factory=lambda: defaultdict(set))

    def __str__(self):
        """
        Export the substitutions, 1 entry per line, as:
        `key` : `counts` : `substitution`
        """
        all_lines = []
        for key in sorted(self.counts.keys()):
            if not bool(self.substitutions[key]):
                continue
            line = [key, str(self.counts[key])]
            line += sorted(self.substitutions[key])
            all_lines.append(self.sep.join(line))

        return "\n".join(all_lines)

    def sub(self, value:str) -> set[str]:
        """ apply a substitution if it exists """
        normed = self.norm_tag(value)
        if normed in self.substitutions:
            return self.substitutions[normed]

        return set([normed])

    def has_sub(self, value):
        return value in self.substitutions

    def update(self, *values):
        for val in values:
            match val:
                case None | "":
                    continue
                case str():
                    self._inc(val)
                case (str() as key, str() as counts):
                    self._inc(key, amnt=int(counts))
                case (str() as key, str() as counts, *subs):
                    norm_key  = self._inc(key, amnt=int(counts))
                    norm_subs = [ self.norm_tag(x) for x in subs]
                    self.substitutions[norm_key].update([x for x in norm_subs if bool(x)])
                case dict():
                    for key, val in val.items():
                        self._inc(key, amnt=val)
                case SubstitutionFile():
                    self.update(val.counts)
                    for tag, subs in val.substitutions.items():
                        self.substitutions[tag].update(subs)
                case TagFile():
                    self.update(val.counts.items())

        return self



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging
printer = logmod.getLogger("doot._printer")

SPIDER  = DootKey.make("spider")
CRAWLER = DootKey.make("crawler")

class RunSpider(Action_p, Importer_M):
    """
      add a spider to the scrapy crawler that is in state
    """

    @DootKey.kwrap.expands("spider")
    @DootKey.kwrap.types("crawler")
    def __call__(self, spec, state, spider, crawler):
        spider_ref   = DootCodeReference.from_str(spider)
        spider_class = spider_ref.try_import()
        deferred     = crawler.crawl(spider_class)

        deferred.addCallback(lambda _: printer.warning("Crawl Complete"))
        printer.info("Crawler Started: %s", deferred)
        return deferred



"""
# template settings for spiders
[spiders.general]
# DUPEFILTER_DEBUG                     = true
JOBDIR                               = ".temp"
FEED_TEMPDIR                         = ".temp"
#LOG_FORMAT                          = ""
# LOG_FILE                             = "scrapy.log"
# disable_existing_loggers = false
# LOG_LEVEL                            = "DEBUG"
# LOG_STDOUT                           = false
REQUEST_FINGERPRINTER_IMPLEMENTATION = "2.7"
TWISTED_REACTOR                      = "twisted.internet.asyncioreactor.AsyncioSelectorReactor"
FEED_EXPORT_ENCODING                 = "utf-8"
USER_AGENT                           = "Scrapy/VERSION (+https://scrapy.org)"
DEPTH_LIMIT                          = 1
CONCURRENT_REQUESTS                  = 16
DOWNLOAD_DELAY                       = 3
DOWNLOAD_TIMEOUT                     = 100
# ROBOTSTXT_OBEY                     = true
#DOWNLOAD_WARNSIZE                   = 33554432
#DOWNLOAD_MAXSIZE                    = 1073741824
#COOKIES_ENABLED                     = false
TELNETCONSOLE_ENABLED                = false

[spiders.throttle]
#AUTOTHROTTLE_ENABLED                = True
#AUTOTHROTTLE_START_DELAY            = 5
#AUTOTHROTTLE_MAX_DELAY              = 60
#AUTOTHROTTLE_TARGET_CONCURRENCY     = 1.0
#AUTOTHROTTLE_DEBUG                  = False

[spiders.caching]
HTTPCACHE_DIR                 = ".temp/cache"
HTTPCACHE_ENABLED             = true
HTTPCACHE_EXPIRATION_SECS     = 0
HTTPCACHE_STORAGE             = "doot.spiders.caching.CacheStorage"
HTTPCACHE_POLICY              = "doot.spiders.caching.CachePolicy"
# HTTPCACHE_IGNORE_HTTP_CODES = []
HTTPCACHE_GZIP                = true

[spiders.pipeline]
# DEFAULT_REQUEST_HEADERS = { "Accept" = "text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8", "Accept-Language"= "en" }
# SPIDER_MIDDLEWARES      = { "doot.spiders.middleware.SpiderMiddleware"     = 543 }
# DOWNLOADER_MIDDLEWARES  = { "doot.spiders.middleware.DownloaderMiddleware" = 543 }
# DOWNLOAD_HANDLERS       = {}
EXTENSIONS                = { "scrapy.extensions.telnet.TelnetConsole"            = false }
# ITEM_PIPELINES          = { "doot.spiders.pipeline.ItemPipeline"  = 300, "doot.spiders.pipeline.SimpleExporter" = 1000 }
"""




#!/usr/bin/env python3
"""

"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
import typing
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1

try:
    # for py 3.10 onwards:
    from typing import TypeAlias
except ImportError:
    TypeAlias = Any

##-- end builtin imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

from collections.abc import Mapping, ItemsView, KeysView, ValuesView
from tomlguard.error import TomlAccessError

super_get = object.__getattribute__
super_set = object.__setattr__

TomlTypes : TypeAlias = str | int | float | bool | list['TomlTypes'] | dict[str,'TomlTypes'] | datetime.datetime

class GuardBase(Mapping[str, TomlTypes]):
    """
    Provides access to toml data (TomlGuard.load(apath))
    but as attributes (data.a.path.in.the.data)
    instead of key access (data['a']['path']['in']['the']['data'])

    while also providing typed, guarded access:
    data.on_fail("test", str | int).a.path.that.may.exist()

    while it can then report missing paths:
    data.report_defaulted() -> ['a.path.that.may.exist.<str|int>']
    """

    _defaulted : ClassVar[set[str]] = set()

    @staticmethod
    def add_defaulted(index:str|list[str], val:TomlTypes, types:str="Any") -> None:
        match index, val:
            case list(), _:
                raise TypeError("Tried to Register a default value with a list index, use a str")
            case str(), bool():
                index_str = f"{index} = {str(val).lower()} # <{types}>"
            case str(), _:
                index_str = f"{index} = {repr(val)} # <{types}>"
            case [*xs], bool():
                index_path = ".".join(xs)
                index_str = f"{index_path} = {str(val).lower()} # <{types}>"
            case [*xs], _:
                index_path = ".".join(xs)
                index_str = f"{index_path} = {val} # <{types}>"
            case _, _:
                raise TypeError("Unexpected Values found: ", val, index)

        GuardBase._defaulted.add(index_str)

    @staticmethod
    def report_defaulted() -> list[str]:
        """
        Report the index paths inject default values
        """
        return list(GuardBase._defaulted)

    def __init__(self, data:dict[str,TomlTypes]=None, *, index:None|list[str]=None, mutable:bool=False):
        super_set(self, "__table", data or {})
        super_set(self, "__index"   , (index or ["<root>"])[:])
        super_set(self, "__mutable" , mutable)

    def __repr__(self) -> str:
        return f"<TomlGuard:{list(self.keys())}>"

    def __setattr__(self, attr:str, value:TomlTypes) -> None:
        if not getattr(self, "__mutable"):
            raise TypeError()
        super_set(self, attr, value)

    def __getattr__(self, attr:str) -> GuardBase | TomlTypes | list[GuardBase]:
        table = self._table()

        if attr not in table and attr.replace("_", "-") not in table:
            index     = self._index() + [attr]
            index_s   = ".".join(index)
            available = " ".join(self.keys())
            raise TomlAccessError(f"{index_s} not found, available: [{available}]")

        match table.get(attr, None) or table.get(attr.replace("_", "-"), None):
            case dict() as result:
                return self.__class__(result, index=self._index() + [attr])
            case list() as result if all(isinstance(x, dict) for x in result):
                index = self._index()
                return [self.__class__(x, index=index[:]) for x in result if isinstance(x, dict)]
            case _ as result:
                return result

    def __getitem__(self, keys:str|list[str]|tuple[str]) -> TomlTypes:
        curr : typing.Self = self
        match keys:
            case tuple():
                for key in keys:
                    curr = curr.__getattr__(key)
            case str():
                curr = self.__getattr__(keys)
            case _:
                pass

        return curr

    def __len__(self) -> int:
        return len(self._table())

    def __call__(self) -> TomlTypes:
        raise TomlAccessError("Don't call a TomlGuard, call a TomlGuardProxy")

    def __iter__(self):
        return iter(getattr(self, "__table").items())

    def __contains__(self, __key: object) -> bool:
        return __key in self.keys()

    def _index(self) -> list[str]:
        return super_get(self, "__index")[:]

    def _table(self) -> dict[str,TomlTypes]:
        return super_get(self, "__table")

    def get(self, key:str, default:TomlTypes|None=None) -> TomlTypes|None:
        if key in self:
            return self[key]

        return default

    def keys(self) -> KeysView[str]:
        # table  = object.__getattribute__(self, "__table")
        table = super_get(self, "__table")
        return table.keys()

    def items(self) -> ItemsView[str, TomlTypes]:
        # match object.__getattribute__(self, "__table"):
        match super_get(self, "__table"):
            case dict() as val:
                return val.items()
            case list() as val:
                return dict({self._index()[-1]: val}).items()
            case GuardBase() as val:
                return val.items()
            case x:
                raise TypeError("Unknown table type", x)

    def values(self) -> ValuesView[TomlTypes]:
        # match object.__getattribute__(self, "__table"):
        match super_get(self, "__table"):
            case dict() as val:
                return val.values()
            case list() as val:
                return val
            case _:
                raise TypeError()



""" Provides the basic Drawing Superclass """
##-- imports
from __future__ import annotations
import logging as root_logger
from dataclasses import InitVar, dataclass, field
from typing import (Any, Callable, ClassVar, Dict, Generic, Iterable, Iterator,
                    List, Mapping, Match, MutableMapping, Optional, Sequence,
                    Set, Tuple, TypeVar, Union, cast)

from ..drawing import draw_circle

##-- end imports

logging = root_logger.getLogger(__name__)

class Drawable:
    """ A Basic Drawable Superclass """

    def __init__(self):
        raise Exception("Drawable Should not be instantiated")

    def draw(self, ctx):
        """ Abstract method that Drawbles implement  """
        raise Exception("Drawble.draw is abstract. Implement it in the calling class")

    def draw_point_cloud(self, ctx, xys, rs, colours):
        """ Draw a collection of points """
        assert(len(xys) == len(rs) == len(colours))
        for (i, a) in enumerate(xys):
            ctx.set_source_rgba(*colours[i])
            draw_circle(ctx, *a, rs[i])



""""
A PyParsing parser to create patterns
"""
##-- imports
from __future__ import annotations

from .pattern_constructor import CTOR_ACT, construct_pattern
from .utils import TimeVar
from enum import Enum
from fractions import Fraction
import pyparsing as pp
import logging as root_logger

##-- end imports

logging = root_logger.getLogger(__name__)
#Base Syntax
s                = pp.Suppress
op               = pp.Optional
opLn             = s(op(pp.lineEnd))
comment          = s(pp.dblSlashComment)

COMMA            = s(pp.Literal(','))
OBRACKET         = s(pp.Literal('['))
CBRACKET         = s(pp.Literal(']'))
COLON            = s(pp.Literal(':'))
QUESTION         = s(pp.Literal('?'))
LESS             = s(pp.Literal('<'))
MORE             = s(pp.Literal('>'))
TILDE            = s(pp.Literal('~'))
DOLLAR           = s(pp.Literal('$'))
VBAR             = s(pp.Literal('|'))

acts             = pp.Or([COMMA, OBRACKET, CBRACKET, COLON, QUESTION, LESS, MORE, TILDE])

NAME             = pp.Word(pp.alphas)
VAR              = DOLLAR + NAME

pattern_contents = pp.OneOrMore(pp.Or([VAR, NAME, acts]))

# Actions
OBRACKET.setParseAction(lambda x: CTOR_ACT.PSTART)
CBRACKET.setParseAction(lambda x: CTOR_ACT.PEND)
COMMA.setParseAction(lambda x: CTOR_ACT.PDUAL)
QUESTION.setParseAction(lambda x: CTOR_ACT.OP)
LESS.setParseAction(lambda x: CTOR_ACT.CSTART)
MORE.setParseAction(lambda x: CTOR_ACT.CEND)
TILDE.setParseAction(lambda x: CTOR_ACT.SIL)

VAR.setParseAction(lambda toks: TimeVar(toks[0]))

def parse_string(s):
    """ The primary access point """
    parsed = pattern_contents.parseString(s)
    # Stack based constructor:
    return construct_pattern(parsed[:])



#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
# import more_itertools as mitz
# from boltons import
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

class EnumBuilder_m:
    """ A Mixin to add a .build(str) method for the enum """

    @classmethod
    def build(cls, val:str) -> Self:
        try:
            match val:
                case str():
                    return cls[val]
                case cls():
                    return val
        except KeyError:
            logging.warning("Can't Create a flag of (%s):%s. Available: %s", cls, val, list(cls.__members__.keys()))

class FlagsBuilder_m:
    """ A Mixin to add a .build(vals) method for EnumFlags """

    @classmethod
    def build(cls, vals:str|list|dict) -> Self:
        match vals:
            case str():
                vals = [vals]
            case list():
                pass
            case dict():
                vals = [x for x,y in vals.items() if bool(y)]

        base = cls.default
        for x in vals:
            try:
                match x:
                    case str():
                        base |= cls[x]
                    case cls():
                        base |= x
            except KeyError:
                logging.warning("Can't create a flag of (%s):%s. Available: %s", cls, x, list(cls.__members__.keys()))

        return base



""""
Red Black Tree Operations, as translated from Cormen.
todo: integrate these into the beachline class
"""
##-- imports
from __future__ import annotations
import logging as root_logger
from .node import Node

##-- end imports

logging = root_logger.getLogger(__name__)

class RBTreeOperations:
    def rotate_left(tree, node):
        """ Rotate the given node left, making the new head be node.right """
        logging.debug("Rotating Left: {}".format(node))
        assert(isinstance(node, Node))
        assert(node.right is not None)
        set_as_root, new_head = node.rotate_left()
        if set_as_root:
            tree.root = new_head
        return new_head

    def rotate_right(tree, node):
        """ Rotate the given node right, making the new head be node.left """
        logging.debug("Rotating Right: {}".format(node))
        assert(isinstance(node, Node))
        assert(node.left is not None)
        set_as_root, new_head = node.rotate_right()
        if set_as_root:
            tree.root = new_head
        return new_head

    def rb_tree_fixup(tree, node):
        """ Verify and fix the RB properties hold """
        while node.parent is not None and node.parent.red:
            parent = node.parent
            grand_parent = parent.parent
            if grand_parent is None:
                break
            elif parent == grand_parent.left:
                y = grand_parent.right
                if y is not None and y.red:
                    parent.red = False
                    y.red = False
                    grand_parent.red = True
                    node = grand_parent
                else:
                    if node == parent.right:
                        node = parent
                        rotate_left(tree, node)#invalidates parent and grandparent
                    node.parent.red = False
                    node.parent.parent.red = True
                    rotate_right(tree, node.parent.parent)
            else:
                y = grand_parent.left
                if y is not None and y.red:
                    parent.red = False
                    y.red = False
                    grand_parent.red = True
                    node = grand_parent
                else:
                    if node == parent.left:
                        node = parent
                        rotate_right(tree, node)#invalidates parent and grandparent
                    node.parent.red = False
                    node.parent.parent.red = True
                    rotate_left(tree, node.parent.parent)
        tree.root.red = False

    def transplant(tree, target, replacement):
        """ Transplant the node replacement, and its subtree,
        in place of node target """
        #logging.debug("Transplanting {} into {}".format(replacement, target))
        if replacement is not None:
            replacement.disconnect_from_parent()

        if target.parent is None:
            logging.debug("Setting root to {}".format(replacement))
            tree.root = replacement
        elif target.parent.on_left(target):
            logging.debug("Transplant linking left")
            parent = target.parent
            target.disconnect_from_parent()
            parent.link_left(replacement)
        else:
            logging.debug("Transplant linking right")
            parent = target.parent
            target.disconnect_from_parent()
            parent.link_right(replacement)

    def rb_tree_delete(tree, node):
        """ Delete a node from the tree and rotate nodes as needed """
        target = node
        target_originally_red = target.red
        current = None
        logging.debug("Deleting Node: {}".format(node))
        if target.left is None:
            logging.debug("No left, transplanting right")
            current = target.right
            transplant(tree, target, target.right)
        elif target.right is None:
            logging.debug("No right, transplanting left")
            current = target.left
            transplant(tree, target, target.left)
        else:
            logging.debug("Both Left and right exist")
            target = target.right.min()
            target_originally_red = target.red
            current = target.right
            if target.parent == node:
                if current is not None:
                    logging.debug("target.parent == node")
                    target.disconnect_from_parent()
                    current.parent = target
            else:
                logging.debug("target.parent != node")
                transplant(tree, target, target.right)
                target.link_right(node.right)

            transplant(tree, node, target)
            target.link_left(node.left)
            target.red = node.red

        if not target_originally_red:
            logging.debug("Fixing up current: {}".format(current))
            rb_delete_fixup(tree, current)

    def rb_delete_fixup(tree, node):
        """ Having deleted a node, rotate remaining nodes
        as necessary to preserve rb properties """
        while node != tree.root and node is not None and not node.red:
            if node.parent.on_left(node):
                w = node.parent.right
                if w is not None and w.red:
                    w.red = False
                    node.parent.red = True
                    rotate_left(tree, node.parent)
                    w = node.parent.right
                if w is not None and (w.left is None or not w.left.red) and \
                (w.right is None or not w.right.red):
                    w.red = True
                    node = node.parent
                elif w is not None:
                    if w.right is None or not w.right.red:
                        w.left.red = False
                        w.red = True
                        rotate_right(tree, w)
                        w = node.parent.right
                    w.red = node.parent.red
                    node.parent.red = False
                    w.right.red = False
                    rotate_left(tree, node.parent)
                    node = tree.root

            else: #mirror for right
                w = node.parent.left
                if w is not None and w.red:
                    w.red = False
                    node.parent.red = True
                    rotate_right(tree, node.parent)
                    w = node.parent.left
                if w is not None and (w.left is None or not w.left.red) and \
                (w.right is None or not w.right.red):
                    w.red = True
                    node = node.parent
                elif w is not None:
                    if w.left is None or not w.left.red:
                        w.right.red = False
                        w.red = True
                        rotate_left(tree, w)
                        w = node.parent.left
                    w.red = node.parent.red
                    node.parent.red = False
                    w.left.red = False
                    rotate_right(tree, node.parent)
                    node = tree.root
            node.red = False
            if w is None:
                node = None



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

printer = logmod.getLogger("doot._printer")

import math
import json
from time import sleep
import sh
import shutil
import jsonlines
import tomlguard as TG

##-- expansion keys
FROM_KEY           : Final[DootKey] = DootKey.make("from")
UPDATE             : Final[DootKey] = DootKey.make("update_")
##-- end expansion keys

class ReadJson(Action_p):
    """
        Read a .json file and add it to the task state as a tomlguard
    """
    _toml_kwargs = [FROM_KEY, UPDATE]

    @DootKey.kwrap.paths("from")
    @DootKey.kwrap.redirects("update_")
    def __call__(self, spec, state, _from, _update):
        if _from.suffix != ".json":
            printer.warning("Read Json expected a .json file, got: %s", _from)

        with open(_from) as fp:
            data     = json.load(fp)
        return { _update : TG.TomlGuard(data) }

class ParseJson(Action_p):
    """ parse a string as json """

    @DootKey.kwrap.types("from")
    @DootKey.kwrap.redirects("update_")
    def __call__(self, spec, state, _from, _update):
        return { _update : json.loads(_from) }

class ReadJsonLines(Action_p):
    """ read a .jsonl file, or some of it, and add it to the task state  """

    @DootKey.kwrap.paths("from")
    @DootKey.kwrap.types("offset", hint={"default":0})
    @DootKey.kwrap.types("count", hint={"default":math.inf})
    @DootKey.kwrap.redirects("update_")
    def __call__(self, spec, state, _from, offset, count, _update):
        if _from.suffix != ".jsonl":
            printer.warning("Read JsonNL expects a .jsonl file, got: %s", _from)

        result = []
        target_end = offset + count
        with jsonlines.open(_from) as reader:
            for i, obj in enumerate(reader):
                if i < offset:
                    continue
                result.append(obj)
                if target_end <= i:
                    break

        return { _update : result }

class WriteJsonLines(Action_p):
    """ Write a list of dicts as a .jsonl file
      optionally gzip the file
    """

    @DootKey.kwrap.types("from")
    @DootKey.kwrap.paths("to")
    def __call__(self, spec, state, _from, _to):
        if _to.suffix != ".jsonl":
            printer.warning("Write Json Lines expected a .jsonl file, got: %s", _to)

        with jsonlines.open(_to, mode='a') as writer:
            writer.write(_from)

class WriteJson(Action_p):
    """ Write a dict as a .json file  """

    @DootKey.kwrap.types("from")
    @DootKey.kwrap.paths("to")
    def __call__(self, spec, state, _from, _to):
        if _to.suffix != ".json":
            printer.warning("Write Json Expected a .json file, got: %s", _to)
        with open(_to, mode='w') as writer:
            json.dump(_from,
                      writer,
                      ensure_ascii=True,
                      indent=4,
                      sort_keys=True)



#!/usr/bin/env python3
"""
Utility actions, such as a debugger

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

import sys
# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
printer = logmod.getLogger("doot._printer")
##-- end logging

import bdb

def action_debugger(spec, state):

    def pstate():
        printer.info("Printing State:")
        printer.info(state)

    def pspec():
        printer.info("Printing Spec:")
        printer.info(spec)

    printer.info("* Entering breakpoint *")
    printer.info("* Call pspec() and pstate() to inspect the spec and state *")
    breakpoint()

    return None

def typecheck(spec, state):
    for key,target_type in spec.kwargs:
        try:
            value = state[key]
            value_type = type(value)
            fullname = "{}:{}".format(value_type.__module__, value_type.__name__)
            if target_type != fullname:
                raise doot.errors.DootActionStateError("Type Error: state.%s : %s != %s", key, fullname, target_type)

            printer.debug("Type Matches: state.%s : %s", key, target_type)

        except (AttributeError, KeyError):
            raise doot.errors.DootActionStateError("State key missing: %s", key)



#!/usr/bin/env python3
"""
Basic Implementations of the handler system protocol
"""
# pylint: disable=abstract-method,invalid-sequence-index,use-a-generator,too-many-lines
# pyright: reportPrivateUsage=warning
##-- imports
from __future__ import annotations

import logging as logmod
from dataclasses import InitVar, dataclass, field, replace
from enum import Enum, EnumMeta
from types import MethodType
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Generic, Iterable,
                    Iterator, Mapping, Match, MutableMapping, NewType,
                    Protocol, Sequence, Tuple, Type, TypeAlias, TypeVar, cast)

##-- end imports

logging                = logmod.getLogger(__name__)

config                 = acab.config
SPACER                 = config.any_of().print.SPACER_SIZE()

DEFAULT_HANDLER_SIGNAL = config.handler.system.DEFAULT_SIGNAL
# TODO import wrapper
Handler                = config.on_fail().imports.specific.handler(wrapper="import"))

PASSTHROUGH         = "_"

# Protocols  ##################################################################

@APE.assert_concrete
class HandlerSystem(HS.HandlerSystem_i):
    """
    The Core Handler system, needing only __call__ and extend to be implemented
    """

    @staticmethod
    def Spec(name:str|Sen_A):
        return HandlerSpec(name)

    def __post_init__(self, init_specs, init_handlers, sieve_fns) -> None:
        sieve_fns     = sieve_fns     or self._default_sieve[:] #type:ignore
        init_specs    = init_specs    or []
        init_handlers = init_handlers or []
        self._quick_sen_fails = set()
        try:
            self.sieve = AcabSieve(sieve_fns)
            self._register_spec(*init_specs)
            self._register_default()
            # add handlers with funcs before structs
            self._register_handler(*sorted(init_handlers, key=lambda x: not x.func))
        except AttributeError as err:
            raise AcabHandlerException("Bad Handler in:", rest=init_handlers) from err

    def __contains__(self, signal) -> bool:
        match signal:
            case str():
                return signal in self.handler_specs
            case Sentence_i():
                return str(signal) in self.handler_specs
            case HS.HandlerSpec_i() | HS.Handler_i():
                return str(signal) in self.handler_specs
            case _ if isinstance(type(signal), EnumMeta):
                return signal.name in self.handler_specs
            case _:
                raise ValueError(f"Unrecognised signal value: {signal}")

    def __bool__(self):
        return 1 < len(self.handler_specs)

    def __len__(self):
        return len(self.handler_specs)

    def __getitem__(self, other):
        if not isinstance(other, str):
            raise ValueError(f"Bad Signal Attempt to HandlerSystem: {other}")
        if other in self.handler_specs:
            return self.handler_specs[other]

        matching_loose = [x for x in self.loose_handlers if x.signal == other]
        if bool(matching_loose):
            return matching_loose[0]

        raise ValueError(f"No Handler or Spec Found for: {other}")

    def __iter__(self):
        for spec in self.handler_specs.values():
            yield spec

    def __repr__(self):
        return f"<{self.__class__.__name__}>"

    def lookup(self, value=None) -> HandlerSpec_A:
        """ run the sieve on the value to get a handler """
        if value is None:
            return self.handler_specs[DEFAULT_HANDLER_SIGNAL]

        is_override    = isinstance(value, HS.HandlerOverride)
        is_passthrough = is_override and value.signal == PASSTHROUGH
        # For using an override to carry data, without an override signal
        if is_passthrough:
            value = value.value

        for key in self.sieve.fifo(value):
            match key:
                case Sentence_i():
                    key_match = self._sub_sen_handler_patch(key)
                    key = str(key)
                case str():
                    key_match   = key in self.handler_specs
                case None:
                    pass
                case _:
                    raise TypeError("Unexpected value returned by sieve", key)

            if is_override and not is_passthrough and not key_match:
                logging.warning(f"Missing Override Handler: {self.__class__} : {key}")
                continue
            if key is None:
                continue

            assert(isinstance(key, str))
            if key_match and bool(self.handler_specs[key]) and self.handler_specs[key].verify(value):
                # TODO return the handler_spec *if* one of its handler's verifies the instruction
                return self.handler_specs[key]

        # Final resort
        return self.handler_specs[DEFAULT_HANDLER_SIGNAL]

    def _sub_sen_handler_patch(self, sen:Sen_A) -> bool:
        """
        Check a handler exists for a.test.sen
        if on isnt found, try a.test
        then just a.

        If a sub sen handler *is* found, patch it
        so the search doesn't have to happen again
        """
        sen_str = str(sen)
        # Try Full
        if sen_str in self:
            return True

        # Check for cached failures
        if sen_str in self._quick_sen_fails:
            return False

        # Try just the last word?
        if str(sen[-1:]) in self:
            new_spec = self.handler_specs[str(sen[-1:])].copy(signal=sen_str, handlers=True)
            self.register(new_spec)
            return True

        # Try subsens
        for index in range(len(sen)-1, 0, -1):
            subsen = str(sen[:index])
            if subsen in self:
                # patch
                logging.debug("Found an applicable sub spec: {} of {}", subsen, sen_str)
                new_spec = self.handler_specs[subsen].copy(signal=sen_str, handlers=True)
                self.register(new_spec)
                return True

        # Else patch in a fast failure
        self._quick_sen_fails.add(sen_str)
        return False

    def override(self, new_signal: bool | str, value, data=None) -> Overrider:
        """ wrap a value to pass data along with it, or explicitly control the signal it produces for handlers """
        data = data or {}

        if isinstance(value, HS.HandlerOverride):
            # override on an override
            return value.replace(signal=new_signal, data=data)

        match new_signal:
            case str() if new_signal in self:
                pass
            case Sentence_i():
                new_signal = str(new_signal)
            case bool() if not new_signal:
                new_signal = PASSTHROUGH
            case _:
                raise TypeError("Bad Override signal", new_signal)

        return HS.HandlerOverride(new_signal, value, data=data)

    #pylint: disable-next=too-many-branches

    def register(self, *others):
        for other in others:
            match other:
                case HandlerFragment_i():
                    for item in other:
                        self.register(item)
                case HS.HandlerSpec_i():
                    self._register_spec(other)
                case HS.Handler_i():
                    self._register_handler(other)
                case HS.HandlerOverride():
                    raise AcabHandlerException("Attempt to register a HandlerOverride, it should be __call__ed instead")
                case dict():
                    self._register_data(other)
                case _:
                    raise AcabHandlerException("Attempt to register unknown type", rest=[other])
        return self

    def _register_default(self):
        if DEFAULT_HANDLER_SIGNAL in self and bool(self.handler_specs[DEFAULT_HANDLER_SIGNAL]):
            return

        self._register_spec(HandlerSpec(DEFAULT_HANDLER_SIGNAL))

    def _register_spec(self, *specs: HandlerSpec_A):
        for spec in specs:
            as_pseudo = str(spec)
            if as_pseudo in self and spec != self.handler_specs[as_pseudo]:
                raise AcabHandlerException(f"Signal Conflict: {spec.signal}") #type:ignore
            if spec not in self:
                self.handler_specs[as_pseudo] = spec.copy(handlers=True)

        # TODO: Then try to register any loose handlers

    def _register_data(self, data: dict[str, Any], *, signal:None|str=None):
        """
        Register additional data that abstractions may access
        """
        if signal is not None:
            raise NotImplementedError()
            # self.handler_specs[signal].require_data(data)

        self._data.update(data)

    def _register_handler(self, *handlers: Handler_A):
        """
        insert a handler into the system, bound to the signal that it listens for
        """
        for handler in handlers:
            if not isinstance(handler, HS.Handler_i):
                raise AcabHandlerException(f"Handler Not Compliant: {handler}", rest=handler)

            as_pseudo = str(handler)
            if as_pseudo not in self:
                self.loose_handlers.append(handler)
            else:
                self.handler_specs[as_pseudo].register(handler)

    def verify_system(self):
        pass

    def extend(self, modules:list[ModuleFragment]) -> None:
        raise NotImplementedError()

    @property
    def signals(self) -> list[str]:
        return list(self.handler_specs.keys())

@APE.assert_concrete
class HandlerSpec(HS.HandlerSpec_i):

    def __str__(self):
        return str(self.signal)

    def __repr__(self):
        return f"<HandlerSpec {self.signal}, flags={self.flags}, func_api={self.func_api}, handlers={len(self.handlers)}>"

    def __bool__(self):
        return bool(self.handlers)

    def __eq__(self, other):
        # TODO handle structs
        # TODO api must be equal
        # TODO data and struct api's must be equal
        if isinstance(other, str):
            return str(self) == other
        if isinstance(other, HandlerSpec):
            return self.signal == other.signal

        return False

    def __len__(self):
        return len(self.handlers)

    def __getitem__(self, i):
        return self.handlers[i].func

    def __call__(self, *args, **kwargs) -> None | Any:
        logging.debug("Calling Handlers of Signal Spec: {}", self.signal)
        # TODO more advanced logic
        args_l : list[Any] = list(args)
        if len(args_l) < 2:
            args_l.append(self.struct)

        result = None
        for handler in self.handlers:
            result = handler(*args_l, **kwargs)

        return result

    def copy(self, *, handlers=False, **kwargs) -> HandlerSpec_A:
        copied = replace(self,
                         func_api=self.func_api,
                         struct_api=self.struct_api,
                         data_api=self.data_api,
                         **kwargs)
        if handlers:
            copied.handlers += self.handlers[:]

        return copied

    # set APIs ################################################################

    def spec_from(self, target):
        """
        Decorator to use as the spec's interface.
        ie:

        @a_spec(signal).from
        class An_ABC:...

        or:

        @a_spec(signal).from
        def a_func...

        """
        # TODO refactor spec_from -> from
        assert(self.func_api is None)
        # assert(isinstance(target, (Type, Callable)))
        self.func_api = target
        return self

    def require_data(self, data: list[str]):
        self.data_api += data
        return self

    def require_struct(self, struct: Type[Structure]):
        assert(self.struct_api is None)
        self.struct_api = struct
        return self

    # Create Handler ##########################################################

    def on(self, *, target=None, **kwargs) -> Handler_A:
        """
        basicly create a Handler.
        The inverse decorator of `from`.
        ie:

        @a_spec.on
        class Implements(An_ABC):...

        or

        @a_spec.on
        def an_implementing_func....
        """
        # TODO check target against spec
        return Handler(self.signal, func=target, **kwargs)

    # Register Handler ########################################################
    #pylint: disable-next=too-many-branches

    def register(self, handler: Handler_A) -> None:
        """ Add a handler into the current, according to the spec instructions
        and handler's flags """
        match handler:
            case HS.Handler_i(struct=None):
                pass
            case HS.Handler_i(struct=struct):
                self.add_struct(handler)

        match handler:
            case HS.Handler_i(func=None):
                pass
            case HS.Handler_i(func=func, flags=flags) if self.flag_e.OVERRIDE in flags:
                self.handlers = [handler]
            case HS.Handler_i(func=func) if HS.Handler_i.__hash__(handler) in self.registered:
                raise HandlerDuplicationException(handler)
            case HS.Handler_i(func=func):
                self.check_api(func=handler.func)
                self.handlers.append(handler)
                self.registered.add(HS.Handler_i.__hash__(handler))

        if self.handler_limit is None:
            return

        # once you hit the minimum number of handlers, you can't go lower
        # NOT ELIF
        if not self.h_limit_history and self.handler_limit.start <= len(self.handlers):
            self.h_limit_history = True

        match (self.h_limit_histoy, len(self.handlers), self.handler_limit.start, self.handler_limit.stop):
            case (history, count, start, _) if count < start:
                raise AcabHandlerException("Handlers Lower Limit Failure")
            case (history, count, _, stop) if stop < count:
                raise AcabHandlerException("Handlers Upper Limit Failure")

    def add_struct(self, handler):
        struct = handler.struct
        if isinstance(struct, type) and isinstance(struct, Structure_i):
            struct = struct.build_default()

        if self.struct is None or self.flag_e.OVERRIDE in handler.flags:
            self.check_api(struct=struct)
            self.struct = struct
        elif self.struct is not None:
            raise AcabHandlerException(f"{self.signal} struct conflict")

    # Check ###################################################################

    def verify(self, instruction) -> bool:
        """
        Check at least 1 handler can accept the instruction
        """
        if not bool(self):
            return False
        return any([x.verify(instruction) for x in self.handlers])

    def check_api(self, *, func=None, data=None, struct=None):
        if func and self.func_api:
            self.check_func_api(func)
        elif data and self.data_api and any([x not in self.data_api for x in data]):
            raise AcabHandlerException("Data api mismatch: {data} against {self.data_api}")
        elif struct is not None and self.struct_api is not None:
            self.check_struct_api(struct)

        return True

    def check_func_api(self, func):
        if self.func_api is None:
            return

        if isinstance(self.func_api, type):
            if any([x not in dir(func) for x in dir(self.func_api)]):
                raise AcabHandlerException(f"Handler Functional API Mismatch: {self.func_api} against {func}")

        elif isinstance(self.func_api, Callable) and not (func.__code__.co_argcount == self.func_api.__code__.co_argcount): # type: ignore[arg-type]
            raise AcabHandlerException(f"Handler Functional Args API Mismatch: {self.func_api} against {func}")

    def check_struct_api(self, struct):
        if self.struct_api is None:
            return

        is_sub = isinstance(struct, type) and issubclass(struct, self.struct_api)
        is_ins = isinstance(struct, self.struct_api)
        is_eq  = self.struct_api == struct

        if not any([is_sub or is_ins or is_eq]):
            raise AcabHandlerException("Struct api mismatch: {struct} against {self.struct_api}")

@APE.assert_concrete(exceptions=['__call__'])
class HandlerComponent(HS.HandlerComponent_i):

    def verify(self, instruction) -> bool:
        return False

    def as_handler(self, *, signal:None|str|Sen_A=None, struct:None|Structure=None, flags:None|list[Enum]=None) -> Handler_A:
        if signal is None:
            signal = self.signal
        assert(signal is not None)
        return Handler(signal,
                       func=self,
                       struct=struct,
                       flags=set(flags or []))



import unittest
import logging
import IPython


class RBTree_Tests(unittest.TestCase):

    def setUp(self):
        self.t = rbtree.RBTree()

    def tearDown(self):
        self.t = None

    #----------
    #creation
    def test_creation(self):
        self.assertIsNotNone(self.t)
        self.assertIsInstance(self.t, rbtree.RBTree)

    #empty
    def test_empty(self):
        self.assertEqual(len(self.t), 0)
        self.assertFalse(bool(self.t))

    #insert
    def test_insert_empty(self):
        self.t.insert(2)
        self.assertEqual(len(self.t), 1)
        self.assertTrue(bool(self.t))
        self.t.insert(3)
        self.assertEqual(len(self.t), 2)

    #min
    def test_min(self):
        self.t.insert(4,2,6,5,2,7,8,4,2,5,2,1)
        self.assertEqual(len(self.t),12)
        m = self.t.min()
        self.assertIsInstance(m, rbtree.Node)
        self.assertEqual(m.value, 1)

    #max
    def test_max(self):
        self.t.insert(4,2,6,5,2,7,8,4,2,5,2,1)
        self.assertEqual(len(self.t),12)
        m = self.t.max()
        self.assertIsInstance(m, rbtree.Node)
        self.assertEqual(m.value, 8)

    def test_cmp(self):
        """ Swaps the ordering using a custom cmp function """
        self.t.cmpFunc = CompFuncs.inverted_comparison
        self.t.insert(4,2,6,5,2,7,8,4,2,5,2,1)
        self.assertEqual(len(self.t),12)
        mi = self.t.min()
        ma = self.t.max()
        self.assertIsInstance(mi, rbtree.Node)
        self.assertIsInstance(ma, rbtree.Node)
        self.assertEqual(mi.value, 1)
        self.assertEqual(ma.value, 8)

    #delete
    def test_delete(self):
        self.t.insert(4,2,6,5,2,7,8,4,2,5,2,1)
        self.assertEqual(len(self.t),12)
        m = self.t.max()
        self.assertIsInstance(m, rbtree.Node)
        self.assertEqual(m.value, 8)
        self.t.delete(m)
        self.assertEqual(len(self.t), 11)

    #search
    def test_search(self):
        self.t.insert(4,2,6,5,2,7,8,4,2,5,2,1)
        self.assertEqual(len(self.t),12)
        found, side = self.t.search(4)
        self.assertIsNotNone(found)
        self.assertIsInstance(found, rbtree.Node)

    def test_search_missing(self):
        self.t.insert(4,2,6,5,2,7,8,4,2,5,2,1)
        self.assertEqual(len(self.t),12)
        found, side = self.t.search(55)
        self.assertIsNone(found)

    def test_get_chain(self):
        baseList = [4,2,6,5,2,7,8,4,2,5,2,1]
        self.t.insert(*baseList)
        self.assertEqual(len(self.t),12)
        chain = self.t.get_chain()
        self.assertIsInstance(chain, list)
        self.assertTrue(all([isinstance(x, rbtree.Node) for x in chain]))
        values = [x.value for x in chain]
        self.assertEqual(values, sorted(baseList))

    def test_colours(self):
        baseList = [4,2,6,5,2,7,8,4,2,5,2,1]
        self.t.insert(*baseList)
        self.assertEqual(len(self.t),12)
        chain = self.t.get_chain()
        for node in chain:
            if node.red:
                self.assertTrue(node.left is None or not node.left.red)
                self.assertTrue(node.right is None or not node.right.red)
        #todo: check it is maintained when inserting and deleting

    #test that the path to farthest leaf is no more than
    #twice as long as path to shortest

if __name__ == "__main__":
    #use python $filename to use this logging setup
      LOGLEVEL = logging.INFO
      logFileName = "log.rbtree_tests"
      logging.basicConfig(filename=logFileName, level=LOGLEVEL, filemode='w')
      console = logging.StreamHandler()
      console.setLevel(logging.WARN)
      logging.getLogger().addHandler(console)
      unittest.main()
      #reminder: user logging.getLogger().setLevel(logging.NOTSET) for log control



""" Voronoi.py : Contains the Voronoi Class, which calculates a graphics independent DCEL
    of a Voronoi Diagram.
"""
##-- imports
from __future__ import annotations
import heapq
import logging as root_logger
import pickle
import sys
from dataclasses import InitVar, dataclass, field
from os.path import isfile
from typing import (Any, Callable, ClassVar, Dict, Generic, Iterable, Iterator,
                    List, Mapping, Match, MutableMapping, Optional, Sequence,
                    Set, Tuple, TypeVar, Union, cast)

import cairo_utils as utils
import numpy as np
import numpy.random as base_random
from cairo_utils import Parabola, rbtree
from cairo_utils.dcel import DCEL, HalfEdge
from cairo_utils.rbtree.comparison_functions import (Directions,
                                                     arc_comparison,
                                                     arc_equality)
from cairo_utils.umath import (bbox_centre, bound_line_in_bbox,
                               get_distance_raw, isClockwise)

from .events import CIRCLE_EVENTS, CircleEvent, SiteEvent, VEvent
from .voronoi_drawing import Voronoi_Debug

##-- end imports

logging = root_logger.getLogger(__name__)

#Constants and defaults
IMAGE_DIR = "imgs"
SAVENAME = "graph_data.pkl"
BBOX = np.array([0, 0, 1, 1]) #the bbox of the final image
EPSILON = sys.float_info.epsilon
MAX_STEPS = 100000
CARTESIAN = True

BASE_VORONOI_VERT_DATA = {"VORONOI_VERTEX" : True}
BASE_VORONOI_EDGE_DATA = {"VORONOI_EDGE"   : True}
BASE_VORONOI_FACE_DATA = {"VORONOI_FAE"    : True}

@dataclass
class VoronoiData:
    """ A Class to construct a voronoi diagram into a given dcel,
    after initial points are added
    in init_graph
    """

    node_size      : int        = field(default=10)
    max_steps      : int        = field(default=MAX_STEPS)
    bbox           : np.ndarray = field(default=BBOX)
    debug_draw     : bool       = field(default=False)
    #File name to pickle data to
    save_file_name : str        = field(default=SAVENAME)
    dcel           : DCEL       = field(default=None)

    current_step   : int                 = field(init=False, default=0)
    #Min Heap of site/circle events
    events         : list[Event]         = field(init=False, default_factory=list)
    #backup of the original sites
    sites          : list[Vertex]        = field(init=False, default_factory=list)
    #backup of all circle events
    circles        : list[Event]         = field(init=False, default_factory=list)
    #storage of breakpoint tuples -> halfedge
    halfedges      : dict[int, HalfEdge] = field(init=False, default_factory=dict)
    #The Beach Line Data Structure
    beachline      : Beachline           = field(init=False, default=None)
    #The sweep line position
    sweep_position : Vertex              = field(init=False, default=None)
    debug          : Voronoi_Debug       = field(init=False, default=None)

    def __post_init__(self, num_of_nodes=10, bbox=BBOX, save_name=SAVENAME, debug_draw=False, n=10, max_steps=MAX_STEPS, dcel=None):
        assert(isinstance(bbox, np.ndarray))
        assert(bbox.shape == (4, ))
        #The output voronoi diagram as a DCEL
        if dcel is None:
            self.__protect_dcel = False
            self.dcel = DCEL(bbox=bbox)
        else:
            self.__protect_dcel = True
            self.dcel = dcel
            if self.dcel.bbox != bbox:
                #todo: check this
                self.bbox = self.bbox

        VEvent.offset = self.bbox[3] - self.bbox[1]

        if self.debug_draw:
            self.debug = Voronoi_Debug(n, IMAGE_DIR, self)

    def reset(self):
        """ Reset the internal data structures """
        if not self.__protect_dcel:
            self.dcel       = DCEL(bbox=self.bbox)
        self.events         = []
        self.circles        = []
        self.halfedges      = {}
        self.sweep_position = None
        self.beachline      = rbtree.RBTree(cmp_func=arc_comparison,
                                            eq_func=arc_equality)

        self.current_step = 0



## base_action.py -*- mode: python -*-
##-- imports
from __future__ import annotations

# import abc
import datetime
# import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
##-- end imports

printer = logmod.getLogger("doot._printer")

from time import sleep
import sys
import sh

class SpeakTimeAction(Action_p):
    """
    A Simple Action that announces the time
    Subclass this and override __call__ for your own actions.
    The arguments of the action are held in self.spec

    """
    _toml_kwargs = ["wait", "background"]
    mac_announce_args = ["-v", "Moira", "-r", "50", "The Time Is "]
    linux_announce_args = ["The Time Is "]
    time_format   = "%H:%M"

    def _current_time(self) -> str:
        now = datetime.datetime.now()
        return now.strftime(self.time_format)

    @DootKey.kwrap.args
    def __call__(self, spec, state, args):
        try:
            match sys.platform:
                case "linux":
                    return self._say_linux(spec, state)
                case "darwin":
                    return self._say_mac(spec, state)
                case _:
                    return False
        except sh.CommandNotFound as err:
            printer.error("Shell Commmand '%s' Not Action: %s", err.args[0], args)
            return False
        except sh.ErrorReturnCode:
            printer.error("Shell Command '%s' exited with code: %s for args: %s", args[0], result.exit_code, args)
            return False


    def _say_linux(self, spec, state:dict):
        cmd    = sh.espeak
        args   = (spec.args or self.mac_announce_args) + [self._current_time()]
        if spec.kwargs.on_fail(False, bool).wait():
            sleep(10)
        result = cmd(*args, _return_cmd=True, _bg=spec.kwargs.on_fail(False, bool).background())
        assert(result.exit_code == 0)
        printer.debug("(%s) Shell Cmd: %s, Args: %s, Result:", result.exit_code, cmd, args)
        printer.info("%s", result, extra={"colour":"reset"})
        return True


    def _say_mac(self, spec, state:dict):
        cmd    = sh.say
        args   = (spec.args or self.mac_announce_args) + [self._current_time()]
        if spec.kwargs.on_fail(False, bool).wait():
            sleep(10)
        result = cmd(*args, _return_cmd=True, _bg=spec.kwargs.on_fail(False, bool).background())
        assert(result.exit_code == 0)
        printer.debug("(%s) Shell Cmd: %s, Args: %s, Result:", result.exit_code, cmd, args)
        printer.info("%s", result, extra={"colour":"reset"})
        return True



"""
Provides the main RBTree class
"""
##-- imports
import logging as root_logger
from dataclasses import InitVar, dataclass, field
from functools import partial
from types import FunctionType
from typing import (Any, Callable, ClassVar, Dict, Generic, Iterable, Iterator,
                    List, Mapping, Match, MutableMapping, Optional, Sequence,
                    Set, Tuple, TypeVar, Union, cast)
##-- end imports

logging = root_logger.getLogger(__name__)

from .comparison_functions import (Directions, default_comparison, default_equality)
from .node import Node
from .operations import rb_tree_delete, rb_tree_fixup

@dataclass
class RBTreeData:
    """ A Red-Black Tree Implementation
    Properties of RBTrees:
    1) Every node is Red Or Black
    2) The root is black
    3) Every leaf is Black, leaves are null nodes
    4) If a node is red, it's children are black
    5) All paths from a node to its leaves contain the same number of black nodes
    """

    cmp_func     : Callable = field(default=default_comparison)
    eq_func      : Callable = field(default=default_equality)
    cleanup_func : Callable = field(default=lambda x: [])

    nodes : List[Node] = field(default_factory=list)
    root : Node = field(default=None)

    def __post_init__(self):
        """ Initialise the rb tree container, ie: the node list """
        #Default Comparison and Equality functions with dummy data ignored
        assert(isinstance(cmp_func, (partial, FunctionType)))
        assert(isinstance(eq_func, (partial, FunctionType)))

    def __len__(self):
        return len(self.nodes)

    def __repr__(self):
        if self.root is None:
            return "<RBTree>"

        return "<RBTree: {}>".format(len(self))



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

import abc
import logging as logmod
import pathlib as pl
from copy import deepcopy
from dataclasses import InitVar, dataclass, field, replace
from re import Pattern
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from uuid import UUID, uuid1
from weakref import ref

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
# If CLI:
# logging = logmod.root
# logging.setLevel(logmod.NOTSET)
##-- end logging

import os
import re
import tomlguard
from jgdv.errors import DirAbsent, LocationExpansionError, LocationError
from jgdv.mixins.path_manip import PathManip_m
from jgdv.enums.location_meta import LocationMeta

KEY_PAT        = doot.constants.patterns.KEY_PATTERN
MAX_EXPANSIONS = doot.constants.patterns.MAX_KEY_EXPANSIONS

class JGDVLocations(PathManip_m):
    """
      A Single point of truth for task access to locations.
      key=value pairs in [[locations]] toml blocks are integrated into it.

      it expands relative paths according to cwd(),
      but can be used as a context manager to expand from a temp different root

      location designations are of the form:
      key = 'location/subdirectory/file'
      simple locations can be accessed as attributes: locs.temp

      more complex locations, with expansions, are accessed as items:
      locs['{temp}/somewhere']
      will expand 'temp' (if it is a registered location)
      """
    locmeta = LocationMeta

    def __init__(self, root:Pl.Path):
        self._root    : pl.Path()               = root.expanduser().absolute()
        self._data    : dict[str, TomlLocation] = dict()
        self._loc_ctx : None|DootLocations      = None

    def __repr__(self):
        keys = ", ".join(iter(self))
        return f"<DootLocations : {str(self.root)} : ({keys})>"

    def __getattr__(self, key) -> pl.Path:
        """
          get a location by name from loaded toml
          delegates to __getitem__
          eg: locs.temp
          """
        if key == "__self__":
            return None
        return self[DootKey.build(key, strict=True)]

    def __getitem__(self, val:str|DootKey|pl.Path|DootTaskArtifact) -> pl.Path:
        """
          eg: doot.locs['{data}/somewhere']
          A public utility method to easily convert paths.
          delegates to DootKey's path expansion

          Get a location using item access for extending a stored path.
          eg: locs['{temp}/imgs/blah.jpg']
        """
        match DootKey.build(val, explicit=True):
            case DootNonKey() as key:
                return key.to_path(locs=self)
            case DootSimpleKey() as key:
                return key.to_path(locs=self)
            case DootMultiKey() as key:
                return key.to_path(locs=self)
            case _:
                raise LocationExpansionError("Unrecognized location expansion argument", val)

    def __contains__(self, key:str|DootKey|pl.Path|DootTaskArtifact):
        """ Test whether a key is a registered location """
        return key in self._data

    def __iter__(self) -> Generator[str]:
        """ Iterate over the registered location names """
        return iter(self._data.keys())

    def __call__(self, new_root=None) -> Self:
        """ Create a copied locations object, with a different root """
        new_obj = DootLocations(new_root or self._root)
        return new_obj.update(self)

    def __enter__(self) -> Any:
        """ replaces the global doot.locs with this locations obj,
        and changes the system root to wherever this locations obj uses as root
        """
        self._loc_ctx = doot.locs
        doot.locs = self
        os.chdir(doot.locs._root)
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback) -> bool:
        """ returns the global state to its original, """
        assert(self._loc_ctx is not None)
        doot.locs     = self._loc_ctx
        os.chdir(doot.locs._root)
        self._loc_ctx = None
        return False

    def get(self, key:DootSimpleKey|str, on_fail:None|str|pl.Path=Any) -> None|pl.Path:
        """
          convert a *simple* key of one value to a path.
          does *not* recursively expand returned paths
          More complex expansion is handled in DootKey and subclasses
        """
        assert(isinstance(key,(DootSimpleKey,str))), (str(key), type(key))
        match key:
            case DootNonKey():
                return pl.Path(key.form)
            case str() | DootSimpleKey() if key in self._data:
                return self._data[key].base
            case _ if on_fail is None:
                return None
            case _ if on_fail != Any:
                return self.get(on_fail)
            case DootSimpleKey():
                return pl.Path(key.form)
            case _:
                return pl.Path(key)

    def normalize(self, path:pl.Path, symlinks:bool=False) -> pl.Path:
        """
          Expand a path to be absolute, taking into account the set doot root.
          resolves symlinks unless symlinks=True
        """
        return self._normalize(path, root=self.root)

    def metacheck(self, key:str|DootKey, meta:LocationMeta) -> bool:
        """ check if any key provided has the applicable meta flags """
        match key:
            case DootNonKey():
                return False
            case DootSimpleKey() if key in self._data:
                return self._data[key].check(meta)
            case DootMultiKey():
                 for k in DootKey.build(key):
                     if k not in self._data:
                         continue
                     if self._data[k].check(meta):
                         return True
            case str():
                return self.metacheck(DootKey.build(key), meta)
        return False

    @property
    def root(self):
        """
          the registered root location
        """
        return self._root

    def update(self, extra:dict|TomlGuard|DootLocations, strict=True) -> Self:
        """
          Update the registered locations with a dict, tomlguard, or other dootlocations obj.
        """
        match extra: # unwrap to just a dict
            case dict():
                pass
            case tomlguard.TomlGuard():
                return self.update(extra._table())
            case DootLocations():
                return self.update(extra._data)
            case _:
                raise doot.errors.LocationError("Tried to update locations with unknown type: %s", extra)

        raw          = dict(self._data.items())
        base_keys    = set(raw.keys())
        new_keys     = set()
        for k,v in extra.items():
            match TomlLocation.build(k, v):
                case _ if k in new_keys and v != raw[k]:
                    raise LocationError("Duplicated, non-matching Key", k)
                case _ if k in base_keys:
                    logging.debug("Skipping Location update of: %s", k)
                    pass
                case TomlLocation() as l if l.check(LocationMeta.normOnLoad):
                    raw[l.key] = TomlLocation.build(k, v, base=self.normalize(l.base))
                    new_keys.add(l.key)
                case TomlLocation() as l:
                    raw[l.key] = l
                    new_keys.add(l.key)
                case _:
                    raise LocationError("Couldn't build a TomlLocation for: (%s : %s)", k, v)

        logging.debug("Registered New Locations: %s", ", ".join(new_keys))
        self._data = raw
        return self

    def ensure(self, *values, task="doot"):
        """ Ensure the values passed in are registered locations,
          error with DirAbsent if they aren't
        """
        missing = set(x for x in values if x not in self)

        if bool(missing):
            raise DirAbsent("Ensured Locations are missing for %s : %s", task, missing)

    def _clear(self):
        self._data = tomlguard.TomlGuard()



#!/usr/bin/env python3
"""

"""
##-- imports
from __future__ import annotations

from dataclasses import replace
import logging as logmod
import warnings
import pathlib
from typing import (Any, Callable, ClassVar, Generic, Iterable, Iterator,
                    Mapping, Match, MutableMapping, Sequence, Tuple, TypeAlias,
                    TypeVar, cast)
##-- end imports

import pytest
from clingo import Control, parse_term, Function, Number

logging = logmod.root

class TestSolver:

    def test_initial(self):
        solver = ClingoSolver()
        self.assert(isinstance(solver, ClingoSolver)

    def test_solver_initialisation(self):
        solver = ClingoSolver()
        assert(solver.ctl is not None)

    def test_solver_with_init_options(self):
        solver = ClingoSolver(options=['-n', 2, '-c', "horizon=2"])
        assert(isinstance(solver, ClingoSolver))

    def test_basic_model(self):
        solver = ClingoSolver("a. b. c. d. e :- a, b, c.")
        result = solver.solve()
        assert(result == 1)

    def test_basic_model_result(self):
        solver = ClingoSolver("a. b. c. d. e :- a, b, c.")
        count = solver.solve()
        assert(count == 1)
        model = solver.results[0]
        assert(isinstance(model,iSolve.InstalModelResult)
        assert(all([str(x) in {"a","b","c","d","e"} for x in model.atoms]))

    def test_basic_fail(self):
        # Leaving off the final `.`
        with pytest.raises(RuntimeError):
            solver = ClingoSolver("a. b. c. d. e :- a, b, c")

    def test_assertion_assignment(self):
        term   = parse_term("testVal")
        term_2 = parse_term("a")
        solver = ClingoSolver("#external testVal. a.")
        count  = solver.solve(["testVal"])
        assert(count == 1)

        assert(term in solver.results[0].atoms)
        assert(term_2 in solver.results[0].atoms)

    def test_assertion_assignment_false(self):
        term   = parse_term("testVal")
        term_2 = parse_term("a")
        solver = ClingoSolver("#external testVal. a.")
        count  = solver.solve()
        assert(count == 1)

        assert(term not in solver.results[0].atoms)
        assert(term_2 in solver.results[0].atoms)

    def test_force_fresh(self):
        term   = parse_term("testVal")
        term_2 = parse_term("a")
        solver = ClingoSolver("#external testVal. a.")
        count  = solver.solve(["testVal"])

        assert(count == 1)
        assert(term in solver.results[0].atoms)
        assert(term_2 in solver.results[0].atoms)
        count2 = solver.solve(fresh=True)
        assert(count == 1)
        assert(len(solver.results) == 1)
        assert(term not in solver.results[0].atoms)
        assert(term_2 in solver.results[0].atoms)

    def test_maintenance_no_change(self):
        term   = parse_term("testVal")
        term_2 = parse_term("a")
        solver = ClingoSolver("#external testVal. a.")
        count  = solver.solve([ term ])

        assert(count == 1)
        assert(term in solver.results[0].atoms)
        assert(term_2 in solver.results[0].atoms)

        count2 = solver.solve()
        assert(len(solver.results) == 2)
        assert(count == 1)
        assert(term in solver.results[-1].atoms)
        assert(term_2 in solver.results[-1].atoms)

    def test_maintenance_with_change(self):
        term         = parse_term("testVal(1)")
        term_2       = parse_term("testVal(2)")
        a_term       = parse_term("a")
        solver = ClingoSolver("#external testVal(1..3). a.")
        count  = solver.solve([term])

        assert(count == 1)
        assert(term in solver.results[0].atoms)
        assert(a_term in solver.results[0].atoms)

        # change the value of the term:
        count2 = solver.solve([term_2, term])
        assert(count == 1)
        assert(str(term) not in solver.results[-1].atoms)
        assert(term_2 in solver.results[-1].atoms)
        assert(a_term in solver.results[-1].atoms)

    def test_maintenance_incremental(self):
        solver = ClingoSolver("""
        #program base.
        on(X, 0) :- init_on(X).
        init_on(a).
        disc(a;b;c).

        #program step(t).
        1 { move(D,t) : disc(D) } 1.

        on(X, t) :- move(X,t).
        """)
        count  = solver.solve()
        assert(count == 1)

        # change the value of the term:
        solver.solve(reground=[("step", [Number(1)])])
        solver.solve(reground=[("step", [Number(5)])])
        solver.solve(reground=[("step", [Number(10)])])
        assert(True)

    def test_all_models(self):
    	# Setup:

        # Pre-check:

        # Action

        # Post-check:

        assert(True)

    @pytest.mark.skip(reason="TODO")
    def test_file_load(self):
        solver = ClingoSolver([])

    @pytest.mark.skip(reason="TODO")
    def test_metadata(self):
        pass




##-- imports
from __future__ import annotations

import logging as logmod
from collections import defaultdict
from importlib.resources import files
from string import Template, ascii_uppercase

##-- end imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

##-- data
tex_path        = files(TEX_loc)

HEADER_PAT      = Template((tex_path / "header_pattern").read_text())
PDF_PRELUDE     = Template((tex_path / "pdf_prelude.tex").read_text())

TRACE           = Template((tex_path / "trace.tex").read_text())

EVENT_MACRO     = Template((tex_path / "event_macro.tex").read_text())
FLUENT_MACRO    = Template((tex_path / "fluent_macro.tex").read_text())

SUB_CHAIN       = Template((tex_path / "subchain.tex").read_text())

CHAIN_NODE      = Template((tex_path / "chain_node.tex").read_text())
CHAIN_LINK      = Template((tex_path / "chain_link.tex").read_text())

FLUENT_SUBCHAIN = Template((tex_path / "fluent_subchain.tex").read_text())
EVENT_SUBCHAIN  = Template((tex_path / "event_subchain.tex").read_text())

FLUENT_LINK     = Template((tex_path / "fluent_link.tex").read_text())
EVENT_LINK      = Template((tex_path / "event_link.tex").read_text())

FLUENT          = Template((tex_path / "fluent.tex").read_text())
INST_TERM       = Template((tex_path / "inst_term.tex").read_text())
TERM            = Template((tex_path / "term.tex").read_text())
TERM_BODY       = Template((tex_path / "term_body.tex").read_text())
##-- end data

class PDFReporter:
    """
        InstalPDFTracer
        Implementation of ABC InstalTracer for pdf output.
    """

    def render_term(self, term, inst=None) -> str:
        """
        Render a passed in ASTTerm into a useable latex string.
        Adapts to handle institutional terms and non.
        """
        body = ""
        if bool(term.params):
            body = self.expand(TERM_BODY,
                               body=",".join(str(x) for x in term.params))

        match inst:
            case None:
                return self.expand(TERM,
                                   head=term.value,
                                   body=body.strip()).strip().replace("_", "\_").strip()
            case _:
                return self.expand(INST_TERM,
                                   head=term.value,
                                   body=body.strip(),
                                   inst=inst).strip().replace("_", "\_").strip()

    def render_fluents(self, state, prior_state=None) -> tuple[set, set, set]:
        """
        Separate out fluents in a state into
        {holding}, {initiated}, and {terminated} sets.

        """
        init   = {self.render_term(x.params[0]) for x in state.fluents if prior_state is None or x not in prior_state}
        holds  = {self.render_term(x.params[0]) for x in prior_state.fluents if x in state} if prior_state is not None else {}
        termin = {self.render_term(x.params[0]) for x in prior_state.fluents if x not in state} if prior_state is not None else {}

        init_render   = {self.expand(FLUENT, mod=BOLD,  term=term).strip() for term in init}
        holds_render  = {self.expand(FLUENT, mod="",    term=term).strip() for term in holds}
        termin_render = {self.expand(FLUENT, mod=SOUT, term=term).strip() for term in termin}

        return sorted(init_render), sorted(holds_render), sorted(termin_render)

    def render_macros(self, trace) -> tuple[list, list]:
        event_macros  = []
        fluent_macros = []
        # Generate Macros
        for i, state, pre, post in trace.contextual_iter():
            macro = macro_key(i)
            # Render events to text
            obs   = [self.render_term(x.params[0]) for x in state.observed]
            occ   = [self.render_term(x.params[0]) for x in state.occurred]
            # Render states to text
            init, holds, term = self.render_fluents(state, pre)
            empty_state       = ["\item" + EMPTY[0]] if not any((init, holds, term)) else []

            event_macros.append(self.expand(EVENT_MACRO,
                                            key=macro,
                                            observed="\\\\\n    ".join(obs or EMPTY),
                                            occurred="\\\\\n    ".join(occ or EMPTY),
                                            )
                                )

            fluent_macros.append(self.expand(FLUENT_MACRO,
                                             key=macro,
                                             holding="\n    ".join(holds or empty_state),
                                             initiated="\n    ".join(init),
                                             terminated="\n    ".join(term),
                                             )
                                 )

        return (event_macros, fluent_macros)

    def trace_to_file(self, trace, path):
        """
        Write to a path the given trace, converted to latex
        """
        self.clear()

        try:
            trace_caption= trace.metadata['filename'].replace("_", "\_")
        except:
            trace_caption = "An Instal Trace"

        # Build the macros
        event_m, fluent_m = self.render_macros(trace)
        # build State chain args
        subchains         = [self.expand(SUB_CHAIN, num=i).strip() for i in range(len(trace))]
        # build State chain nodes
        chain_nodes       = [self.expand(CHAIN_NODE, num=i).strip() for i in range(len(trace))]
        # link the main states together
        chain_links       = [self.expand(CHAIN_LINK, num=i, numPlus=i+1).strip() for i in range(len(trace)-1)]

        # build event subchains
        event_subchains   = [self.expand(EVENT_SUBCHAIN, key=macro_key(i), num=i).strip() for i in range(len(trace)-1)]
        # build fluent subchains
        fluent_subchains  = [self.expand(FLUENT_SUBCHAIN, num=i, key=macro_key(i)) for i in range(len(trace))]
        # link event subchains to main
        event_links       = [self.expand(EVENT_LINK, num=i).strip() for i in range(len(trace)-1)]
        # link fluent subchains to main
        fluent_links      = [self.expand(FLUENT_LINK, num=i).strip() for i in range(len(trace))]

        # Trace
        trace_template = TRACE

        is_partial = True if trace[0].timestep != 0 else False
        is_vertical = False

        trace = self.expand(trace_template,
                            is_partial=r"\IsPartialTracetrue" if is_partial else "",
                            is_vertical=r"\IsVerticalTracetrue" if is_vertical else "",
                            last=len(trace)-1,
                            subchains="\n".join(subchains),
                            chain_nodes="\n".join(chain_nodes),
                            chain_links="\n".join(chain_links),
                            fluent_subchains="\n".join(fluent_subchains),
                            event_subchains="\n".join(event_subchains),
                            fluent_links="\n".join(fluent_links),
                            event_links="\n".join(event_links))

        # Combine it all in the prelude:
        self.insert(PDF_PRELUDE,
                    event_macros="\n".join(event_m),
                    fluent_macros="\n".join(fluent_m),
                    trace=trace,
                    caption=trace_caption
                    )

        # write
        with open(path, 'w') as f:
            f.write("\n".join(self._compiled_text))



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

TAG_NORM : Final[re.Pattern] = re.compile(" +")

@dataclass
class TagFile:
    """ A Basic TagFile holds the counts for each tag use """

    counts : dict[str, int] = field(default_factory=lambda: defaultdict(lambda: 0))
    sep    : str            = field(default=" : ")
    ext    : str            = field(default=".tags")

    norm_regex : re.Pattern  = TAG_NORM

    @classmethod
    def read(cls, fpath:pl.Path, sep=None) -> TagFile:
        obj = cls(sep=sep or cls.sep)
        for i, line in enumerate(fpath.read_text().split("\n")):
            try:
                obj.update(tuple(x.strip() for x in line.split(obj.sep)))
            except Exception as err:
                logging.warning("Failure Tag Reading %s (l:%s) : %s : %s", fpath, i, err, line)

        return obj

    def __iter__(self):
        return iter(self.counts)

    def __str__(self):
        """
        Export the counts, 1 entry per line, as:
        `key` : `value`
        """
        all_lines = []
        for key in sorted(self.counts.keys(), key=lambda x: x.lower()):
            if not bool(self.counts[key]):
                continue
            all_lines.append(self.sep.join([key, str(self.counts[key])]))
        return "\n".join(all_lines)

    def __repr__(self):
        return f"<{self.__class__.__name__}: {len(self)}>"

    def __iadd__(self, values):
        return self.update(values)

    def __len__(self):
        return len(self.counts)

    def __contains__(self, value):
        return self.norm_tag(value) in self.counts

    def _inc(self, key, *, amnt=1):
        norm_key = self.norm_tag(key)
        self.counts[norm_key] += amnt
        return norm_key

    def update(self, *values):
        for val in values:
            match val:
                case None | "":
                    continue
                case str():
                    self._inc(val)
                case [str() as key]:
                    self._inc(key)
                case (str() as key, str() as counts):
                    self._inc(key, amnt=int(counts))
                case TagFile():
                    self.update(*values.counts.items())
                case set():
                    self.update(*val)
        return self

    def to_set(self) -> Set[str]:
        return set(self.counts.keys())

    def get_count(self, tag):
        return self.counts[self.norm_tag(tag)]

    def norm_tag(self, tag):
        return self.norm_regex.sub("_", tag.strip())



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

from time import sleep
import timeit
import time
from random import random

autorange_fmt : Final[str] = "%-*10s : %-*5d calls took: %-*8.2f seconds"
result_fmt    : Final[str] = "Attempt %-*5d : %-*8.2f seconds"
block_fmt     : Final[str] = "%-*10s : %-*8.2f seconds"
once_fmt      : Final[str] = "%-*10s : %-*8.2f seconds"

class JGDVTimer:
    """ Utility Class to time code execution.

      see https://docs.python.org/3/library/timeit.html
    """

    def __init__(self, count=10, repeat=5, keep_gc=False, group:None|str=None):
        self.level            = level
        self.count            = count
        self.repeat           = repeat
        self.keep_gc          = keep_gc
        self.group : str      = f"{group}::" if group else ""
        self.total            = 1.0
        self.once_log         = []

    def msg(self, str, *args):
        logging.debug(str, *args)

    def _set_name(self, name, stmt):
        match name, stmt:
            case str(), _:
                self.current_name = self.group + name
            case _:
                self.current_name = self.group + stmt.__qualname__

    def autorange_cb(self, number, took):
        self.msg(autorange_fmt, self.current_name, number, took)
        self.total += took

    def auto(self, stmt, name=None):
        self._set_name(name, stmt)
        self.msg("-- Autoranging: %s", self.current_name")
        timer = timeit.Timer(stmt, globals=globals())
        timer.autorange(self.autorange_cb)

    def repeats(self, stmt, name=None):
        self._set_name(name, stmt)
        self.msg("-- Repeating %s : Timing %s repeats of %s trials", self.current_name, self.repeat, self.count)
        timer  = timeit.Timer(stmt, globals=globals())
        results = timer.repeat(repeat=self.repeat, number=self.count)
        for i, result in enumerate(results):
            self.msg(result_fmt, i, result)

    def block(self, stmt, name=None):
        self._set_name(name, stmt)
        self.msg("-- Running Block %s : Timing block of %-*5f trials", self.current_name, self.count)
        timer  = timeit.Timer(stmt, globals=globals())
        result = timer.timeit(self.count)
        self.msg(block_fmt, self.current_name, result)

    def once(self, stmt, name=None):
        self._set_name(name, stmt)
        self.msg("-- Running Call Once: %s", self.current_name)
        timer  = timeit.Timer(stmt, globals=globals())
        result = timer.timeit(1)
        self.once_log.append((self.current_name, result))
        self.msg(once_fmt, self.current_name, result)

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        self.msg("-- Finished %s : %-*8.2f", self.group, self.total)
        if self.once_log:
            self.msg("-- Largest Single Call: %s", max(self.once_log, key=lambda x: x[1]))

        return



"""
Commands for the REPL

"""
##-- imports
from __future__ import annotations

import importlib
import logging as logmod
import re
import traceback
from datetime import datetime
from enum import Enum
from os.path import abspath, exists, expanduser, split, splitext

##-- end imports

logging = logmod.getLogger(__name__)

import pyparsing as pp

try:
    import readline
except ImportError:
    readline = None

config  = acab.config

@register
def do_prompt(self, line):
    """
    Change the prompt of the repl

    Usage:
    prompt {newprompt}
    """
    self.state.prompt = line.strip()

@register
def do_multi(self, line):
    """
    Activate multi-line collation

    """
    if not self.state.in_multi_line:
        # Start
        logging.info("Activating multi line")
        self.state.in_multi_line = True
        self.state.collect_str = []
        self.state.prompt_bkup = self.state.prompt
        self.state.prompt = self.state.prompt_ml
        self.state.indent = 0
        if bool(readline):
            # indent modification based on:
            # https://stackoverflow.com/questions/8505163

            def input_hook():
                indent_str = self.state.indent * "    "
                readline.insert_text(indent_str)
                readline.redisplay()

            readline.set_pre_input_hook(input_hook)
    else:
        logging.info("Deactivating multi line")
        collected = "\n".join(self.state.collect_str)
        self.state.in_multi_line = False
        self.state.prompt = self.state.prompt_bkup
        self.state.indent = 0
        if bool(readline):
            readline.set_pre_input_hook()

        logging.info(f"Collected: {collected}")
        if bool(line):
            self.onecmd(line + " " + collected)
        else:
            self.onecmd(collected)

@register
def do_pop(self, line):
    """
    Pop off the last string added in multi-line mode,
    for when an error was made
    """
    self.state.collect_str.pop()
    logging.info(f"Collecting: {self.state.collect_str}")

@register
def do_collect(self, line):
    """ Add a line to the multi line collection,
    ready to be used as one statement when multi line is closed """
    assert(self.state.in_multi_line)
    if line.strip() == "end":
        self.state.indent = max(self.state.indent - 1, 0)

    curr_indent = self.state.indent
    curr_indent_str = curr_indent * "    "

    self.state.collect_str.append(curr_indent_str + line)
    if line[-1] == ":":
        self.state.indent += 1

    logging.info("Collecting: {}".format(self.state.collect_str))

@register
def do_echo(self, line):
    """
    Toggle echoing of working memory state
    """
    self.state.echo = not self.state.echo

@register
def do_suppress(self, line):
    """
    Repl Command to suppress all repl commands
    (except itself and `exit`.)
    """
    # TODO
    return



"""
The how a parser behaves in Instal, and a utility to test parsing
"""
##-- imports
from __future__ import annotations

import abc
import logging as logmod
import pathlib as pl
from dataclasses import InitVar, dataclass, field
from importlib.readers import MultiplexedPath
from types import NoneType
from typing import (IO, TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, List, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable)
from unittest.util import safe_repr

##-- end imports

logging = logmod.getLogger(__name__)

from jgdv.apis.clingo.ast import SolverAST
import pyparsing as pp
import pyparsing.testing as ppt

class SolverDSL_i(metaclass=abc.ABCMeta):
    """
    The abstract api of an Instal Parser.

    default implementations use pyparsing,
    but that isn't necessary.

    """

    @abc.abstractmethod
    def parse(self, text:str|pl.Path, *, parser_source=str|pl.Path) -> list[SolverAST]: pass


class SolverDSLTestCase(TestCase):
    """
    A Utility class for simplifying testing parsers
    """

    current_parse_text : None|str = None

    @classmethod
    def setUpClass(cls):
        # pylint: disable=consider-using-f-string
        LOGLEVEL      = logmod.DEBUG
        LOG_FILE_NAME = "log.{}".format(pl.Path(__file__).stem)

        cls.file_h        = logmod.FileHandler(LOG_FILE_NAME, mode="w")
        cls.file_h.setLevel(LOGLEVEL)

        logging.root.addHandler(cls.file_h)
        logging.root.setLevel(logmod.NOTSET)

    @classmethod
    def tearDownClass(cls):
        logmod.root.removeHandler(cls.file_h)

    def assertFilesParse(self, dsl:pp.ParserElement, *files:str|pl.Path, loc:None|MultiplexedPath=None):
        """
        Assert all files parse without error.
        Can be simple names all appended to the path `loc` if provided,
        which is expected to be a MultiplexedPath provided by
        importlib.resources.files
        """
        self.assertIsInstance(dsl, pp.ParserElement)
        self.assertTrue(isinstance(loc, (NoneType, MultiplexedPath)))
        file_paths = (pl.Path(path) for path in files)
        file_locs  = (loc / path for path in file_paths) if loc is not None else file_paths

        for path in file_locs:
            with self.subTest(msg=path):
                self.assertTrue(path.exists())
                try:
                    dsl.parse_file(path, parse_all=True)
                except pp.ParseException as err:
                    raise self.failureException("\n"+err.explain(0)) from None

    def yieldParseResults(self, dsl:pp.ParserElement, *tests) -> Iterator[Any]:
        """
        For each test, yield its result and additional values
        for manual testing
        """
        for test in tests:
            text, data = None, None
            self.assertIsInstance(test, (str, tuple, dict))
            match test:
                case str():
                    text = test
                    data = None
                case tuple():
                    text = test[0]
                    data = test
                case dict():
                    self.assertIn("text", test)
                    text = test['text']
                    data = test
                case _:
                    self.failureException("Test passed to yieldParseResults is confusing: %s", test)

            self._set_current_parse_text(text)
            try:
                result   = dsl.parse_string(text, parse_all=True)
            except pp.ParseException as err:
                raise self.failureException("\n"+err.explain(0)) from None

            yield result, data

        self._clear_current_parse_text()

    def assertParseResults(self, dsl:pp.ParserElement, *tests):
        """
        Run Tests of definition (testStr, {namedresults}?, listResults...)
        """
        self.assertIsInstance(dsl, pp.ParserElement)
        for test in tests:
            self.assertIsInstance(test, tuple, test[0])
            self.assertGreaterEqual(len(test), 2, test[0])
            with self.subTest(msg=test[0]):
                try:
                    result   = dsl.parse_string(test[0], parse_all=True)
                except pp.ParseException as err:
                    raise self.failureException("\n"+err.explain(0)) from None

                named    = test[1] if isinstance(test[1], dict) else {}
                expected = test[1 if not bool(named) else 2:]

                for x,y in named.items():
                    self.assertIn(x, result, test[0])
                    self.assertEqual(result[x], y, test[0])

                self.assertEqual(len(result), len(expected), test[0])
                for x,y in zip(result, expected):
                    self.assertEqual(x, y, test[0])

    def assertParseResultsIsInstance(self, dsl:pp.ParserElement, *tests):
        """
        Run Tests of definition (testStr, {namedresults}?, listResults...)
        """
        self.assertIsInstance(dsl, pp.ParserElement)
        for test in tests:
            self.assertIsInstance(test, tuple)
            self.assertGreaterEqual(len(test), 2)
            with self.subTest(test[0]):
                result   = dsl.parse_string(test[0], parse_all=True)
                expected = test[1:]
                self.assertEqual(len(result), len(expected))
                self.assertTrue(all(isinstance(x, type) for x in expected))

                for x,y in zip(result, expected):
                    self.assertIsInstance(x, y)

    def assertParserFails(self, dsl:pp.ParserElement, *tests):
        """
        Run Tests expected to fail: (testStr, failLoc)
        """
        self.assertIsInstance(dsl, pp.ParserElement)
        for test in tests:
            with self.subTest(test[0]):
                fail_loc = test[1]
                test_exc = test[2] if len(test) > 2 else pp.ParseException
                self.assertIsInstance(fail_loc, int)
                self.assertLess(fail_loc, len(test[0]))

                with self.assertRaises(test_exc) as cm:
                    dsl.parse_string(test[0], parse_all=True)

                exc = cm.exception
                if hasattr(exc, "loc"):
                    self.assertEqual(exc.loc, fail_loc,
                                     "\n"+exc.explain(0))

    def assertAllIn(self, values, container):
        for value in values:
            self.assertIn(value, container)

    def _formatMessage(self, msg, standardMsg):
        """Honour the longMessage attribute when generating failure messages.
        If longMessage is False this means:
        * Use only an explicit message if it is provided
        * Otherwise use the standard message for the assert

        If longMessage is True:
        * Use the standard message
        * If an explicit message is provided, plus ' : ' and the explicit message
        """
        # pylint: disable=too-many-return-statements, consider-using-f-string
        if not self.longMessage:
            return msg or standardMsg

        if msg is None and self.current_parse_text is None:
            return standardMsg

        if msg is None and self.current_parse_text is not None:
            return standardMsg + f'\n\n in:\n{self.current_parse_text}'

        try:
            # don't switch to '{}' formatting in Python 2.X
            # it changes the way unicode input is handled
            if self.current_parse_text:
                return '%s : %s\n\nin:\n%s' % (standardMsg, msg, self.current_parse_text)

            return '%s : %s' % (standardMsg, msg)
        except UnicodeDecodeError:
            if self.current_parse_text is not None:
                return  '%s : %s\n\nin:\n%s' % (safe_repr(standardMsg), safe_repr(msg), safe_repr(self.current_parse_text))

            return  '%s : %s' % (safe_repr(standardMsg), safe_repr(msg or self.current_parse_text))

    def _set_current_parse_text(self, text:str):
        as_lines = text.split("\n")
        rejoined = "\n".join(x.strip() for x in as_lines)
        self.current_parse_text = f'"{rejoined}"\n(ws trimmed)'

    def _clear_current_parse_text(self):
        self.current_parse_text = None



"""
Module of different easing functions
All functions take xs from -1 - 1,
and produce ys from 0 - 1
"""
##-- imports
from __future__ import annotations

from enum import Enum
import numpy as np
from .constants import PI

##-- end imports

CODOMAIN = Enum("Codomain of the curve", "FULL LEFT RIGHT")

def quantize(xs, r=None, q=5, bins=None):
    """ Given an array of numbers, quantize into a number of bins """
    if r is None:
        r = [xs.min(), xs.max()]
    if bins is None:
        bins = np.linspace(r[0], r[1], q)
    inds = np.digitize(xs, bins, right=True)
    return np.array([bins[x] for x in inds])

def interp(xs, r=None, codomain_e=CODOMAIN.FULL):
    """ Auto interpolate input to an -1 to 1 scale,
    r can override the calculated input domain of xs.min/max"""
    codomain = [-1, 1]
    if codomain_e == CODOMAIN.LEFT:
        codomain = [-1, 0]
    elif codomain_e == CODOMAIN.RIGHT:
        codomain = [0, 1]
    if r is None:
        r = [xs.min(), xs.max()]
    return np.interp(xs, r, codomain)

def pow_abs_curve(xs, a, r=None, codomain_e=CODOMAIN.FULL):
    """ 1 - pow(abs(x), a)
    Var a changes the curve.
    0.5: exp tri
    3.5: saturated bell """
    assert(0.5 <= a <= 3.5)
    ixs = interp(xs, r=r, codomain_e=codomain_e)
    return 1 - pow(np.abs(ixs), a)

def pow_cos_pi(xs, a, r=None, codomain_e=CODOMAIN.FULL):
    """ pow(cos(pi * x / 2) a)
    Var a changes curve
    0.5: semi circle
    3.5: gaussian window """
    assert(0.5 <= a <= 3.5)
    ixs = interp(xs, r=r, codomain_e=codomain_e)
    return pow(np.cos(PI * ixs * 0.5), a)

def pow_abs_sin_pi(xs, a, r=None, codomain_e=CODOMAIN.FULL):
    """ 1 - pow(abs(sin(pi * x / 2)), a)
    Var a changes curve
    0.5: exp_tri
    3.5: slightly saturated f
    """
    assert(0.5 <= a <= 3.5)
    ixs = interp(xs, r=r, codomain_e=codomain_e)
    return 1 - pow(np.abs(np.sin(PI * ixs * 0.5)), a)

def pow_min_cos_pi(xs, a, r=None, codomain_e=CODOMAIN.FULL):
    """ 1 - pow(min(cos(pi * x / 2), 1 - abs(x)), a)
    Var a changes curve
    0.5: leaf
    3.5: ep_tri
    """
    assert(0.5 <= a <= 3.5)
    ixs = interp(xs, r=r, codomain_e=codomain_e)
    ixs_balanced = np.column_stack((np.cos(PI * ixs * 0.5), 1 - np.abs(ixs)))
    return 1 - pow(np.min(ixs_balanced, axis=1), a)

def pow_max_abs(xs, a, r=None, codomain_e=CODOMAIN.FULL):
    """ 1 - pow(max(0, abs(x)) * 2 - 1, a)
    Var a changes curve
    0.5: exp_tri
    3.5: slightly saturated f
    """
    assert(0.5 <= a <= 3.5)
    ixs = interp(xs, r=r, codomain_e=codomain_e)
    ixs_zero = np.column_stack((np.abs(ixs) * 2 - 1, np.zeros(len(ixs))))
    return 1 - pow(np.max(ixs_zero, axis=1), a)

def sigmoid(xs, a, r=None, codomain_e=CODOMAIN.FULL):
    """ 1 / (1 + e^(-5 * x)
    Var a does nothing
    """
    ixs = interp(xs, r=r, codomain_e=codomain_e)
    return 1 / (1 + pow(np.e, -5 * ixs))

def linear(xs, a, r=None, codomain_e=CODOMAIN.FULL):
    """ A linear easing: x = y """
    ixs = interp(xs, r=r, codomain_e=codomain_e)
    return ixs

def static(xs, a, r=None, codomain_e=CODOMAIN.FULL):
    """ No matter the values passed in return an array of 1's """
    return np.ones(xs.shape)

def soft_knee(i, threshold, ratio, knee):
    # A Simple soft_knee compression curve
    # Source: https://se.mathworks.com/help/audio/ref/compressor-class.html
    # Input: np.array, Threshold, Ratio, Knee width
    kneeStart, kneeEnd = (threshold - knee/2, threshold + knee/2)
    under              = i < kneeStart
    over               = kneeEnd < i
    inKnee             = np.invert(under) * np.invert(over)

    k_f              = (1/ratio - 1)
    intermediate     = (i - kneeEnd)
    intermediate_pow = pow(intermediate,2)
    k_div_amnt       = 2 * knee
    k_mod            = (k_f * intermediate_pow) / k_div_amnt
    k_red            = i + k_mod

    over_red = threshold + ((i - threshold)) / ratio

    return (i * under) + (k_red * inKnee) + (i * over_red)


def soft_knee_2(i,t,r,k):
    # A Simple soft_knee compression curve
    # Source: https://se.mathworks.com/help/audio/ref/compressor-class.html
    # Input: 1d np.array(), Threshold, Ratio, Knee width
    under = np.array([x for x in i if x < (t - k/2)])
    inKnee = np.array([x for x in i if (t - k/2) <= x and x <= (t + k/2)])
    over = np.array([x for x in i if (t + k/2) < x])

    k_f = (1/r - 1)
    intermediate = inKnee - t + k/2
    intermediate_pow = pow(intermediate,2)
    k_div_amnt = 2 * k
    k_mod = (k_f * intermediate_pow) / k_div_amnt
    k_red = inKnee + k_mod

    over_red = t + ((over - t)) / r

    return np.concatenate((under,k_red,over_red))


def hard_knee(i,t,r):
    """
    A Simple hard knee transfer function
    Input: np.array(), 1d,
    Threshold
    Ratio
    """
    under = np.array([x for x in i if x < t])
    over = np.array([x for x in i if x >= t])
    compressed = t + ((over - t) / R)
    return np.concatenate((under,compressed))


def shape(x):
    #An example shaping function
    return (4/9 * pow(x,6)) - (17/9 * pow(x,4)) + (22/9 * pow(x,2))



import unittest
import logging
t = time.Time

class TestTime(unittest.TestCase):

    def setUp(self):
        return 1

    def tearDown(self):
        return 1

    #----------
    # ARC TESTS
    def test_arc_creation(self):
        an_arc = Arc(t(0,1), t(1,1))
        self.assertIsNotNone(an_arc)

    def test_arc_contains_true(self):
        an_arc = Arc(t(0,1), t(1,1))
        self.assertTrue(t(1,4) in an_arc)
        self.assertTrue(t(3,4) in an_arc)

    def test_arc_contains_false(self):
        an_arc = Arc(t(0,1), t(1,2))
        self.assertFalse(t(3,4) in an_arc)

    #get the size from the arc
    def test_arc_size(self):
        an_arc = Arc(t(1,4), t(1,2))
        self.assertEqual(an_arc.size(), t(1,4))

    def test_arc_equality(self):
        an_arc = Arc(t(1,4), t(2,3))
        an_arc_2 = Arc(t(1,4), t(2,3))
        self.assertEqual(an_arc, an_arc_2)

    def test_arc_equality_failure(self):
        an_arc = Arc(t(1,4), t(2,3))
        an_arc_2 = Arc(t(2,4), t(2,3))
        self.assertNotEqual(an_arc, an_arc_2)

    def test_arc_bound(self):
        an_arc = Arc(t(1,4), t(1,2))
        another_arc = Arc(t(1,3), t(2,3))
        combined = an_arc.bound(another_arc)
        combined_other = another_arc.bound(an_arc)
        self.assertEqual(combined, combined_other)
        self.assertEqual(combined.start, t(1,4))
        self.assertEqual(combined.end, t(2,3))


    #--------------------
    # EVENT TESTS
    def test_event_creation(self):
        anEvent = Event(Arc(t(0,1), t(1,1)), "a")
        self.assertIsNotNone(anEvent)

    #call an event
    def test_event_call(self):
        anEvent = Event(Arc(t(0,1), t(1,1)), "a")
        callResult = anEvent(t(1,2))
        self.assertEqual(len(callResult), 1)

    def test_event_call_outside_range(self):
        anEvent = Event(Arc(t(0,1), t(1,1)), "a")
        callResult = anEvent(t(2,1))
        self.assertEqual(len(callResult), 0)

    #call an event that holds a pattern
    def test_event_call_pattern(self):
        aPattern = time.Pattern(Arc(t(0,1),t(1,1)),
                           [ Event(Arc(t(0,1),t(1,2)), "a"),
                             Event(Arc(t(1,2),t(1,1)), "b"),
                           ])
        anEvent = Event(Arc(t(0,1), t(1,1)), aPattern, True)
        callResult = anEvent(t(1,2))
        self.assertEqual(len(callResult), 1)
        self.assertEqual(callResult[0].values, "b")

    #get the base set
    def test_event_base(self):
        anEvent = Event(Arc(t(0,1),t(1,1)), "a")
        base = anEvent.base()
        self.assertEqual(len(base), 2)

    def test_event_base_pattern(self):
        aPattern = time.Pattern(Arc(t(0,1),t(1,1)),
                                [ Event(Arc(t(1,4),t(1,2)), "a"),
                                  Event(Arc(t(1,6),t(3,8)), "b") ])
        anEvent = Event(Arc(t(0,1),t(1,1)), aPattern, True)
        base = anEvent.base()
        self.assertEqual(len(base), 6)

    #get the key
    def test_event_get_key(self):
        anEvent = Event(Arc(t(0,1),t(1,1)), "a")
        key = anEvent.key()
        self.assertEqual(key, t(0,1))

    #sort by key
    def test_event_sort_by_key(self):
        events = [ Event(Arc(t(1,2),t(1,1)), "a"),
                   Event(Arc(t(0,1),t(1,2)), "b"),
                   Event(Arc(t(2,1),t(3,1)), "c")]
        sorted_events = sorted(events, key=lambda x: x.key())
        values = [x.values for x in sorted_events]
        self.assertEqual(values, ["b","a","c"])

    #check contains
    def test_event_contains(self):
        anEvent = Event(Arc(t(0,1),t(1,2)), "a")
        self.assertTrue(t(1,4) in anEvent)
        self.assertFalse(t(3,4) in anEvent)


    #--------------------
    # PATTERN TESTS
    def test_pattern_creation(self):
        aPattern = time.Pattern(Arc(t(0,1), t(1,1)), [])
        self.assertIsNotNone(aPattern)

    #call empty
    def test_pattern_call_empty(self):
        aPattern = time.Pattern(Arc(t(0,1),t(1,1)), [])
        self.assertEqual(len(aPattern(t(0,1))), 0)

    #call with events
    def test_pattern_call(self):
        aPattern = time.Pattern(Arc(t(0,1), t(1,1)),
                                [ Event(Arc(t(0,1),t(1,2)), "a"),
                                  Event(Arc(t(1,2),t(1,1)), "b")])
        res = aPattern(t(1,2))
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0].values, "b")

    #call with patterns
    def test_pattern_call_with_internal_pattern_start(self):
        aPattern = time.Pattern(Arc(t(0,1), t(1,2)),
                                [ Event(Arc(t(0,1), t(1,2)), "a"),
                                  Event(Arc(t(1,2),t(1,1)), "b") ])

        mainPattern = time.Pattern(Arc(t(0,1), t(1,1)),
                                   [ Event(Arc(t(0,1), t(1,1)), "c"),
                                     aPattern ])

        result = mainPattern(t(0,1), True)
        self.assertEqual(len(result), 2)
        self.assertTrue("a" in result)
        self.assertTrue("c" in result)
        self.assertFalse("b" in result)

    def test_pattern_call_with_internal_pattern_scale(self):
        aPattern = time.Pattern(Arc(t(0,1), t(1,2)),
                                [ Event(Arc(t(0,1), t(1,2)), "a"),
                                  Event(Arc(t(1,2),t(1,1)), "b") ])

        mainPattern = time.Pattern(Arc(t(0,1), t(1,1)),
                                   [ Event(Arc(t(0,1), t(1,1)), "c"),
                                     aPattern ])

        #checks the internal pattern is scaled appropriately:
        result = mainPattern(t(1,4), True)
        self.assertEqual(len(result), 2)
        self.assertFalse("a" in result)
        self.assertTrue("b" in result)
        self.assertTrue("c" in result)

    def test_pattern_call_with_internal_pattern_end(self):
        aPattern = time.Pattern(Arc(t(0,1), t(1,2)),
                                [ Event(Arc(t(0,1), t(1,2)), "a"),
                                  Event(Arc(t(1,2),t(1,1)), "b") ])

        mainPattern = time.Pattern(Arc(t(0,1), t(1,1)),
                                   [ Event(Arc(t(0,1), t(1,1)), "c"),
                                     aPattern ])

        #checks the internal pattern ends appropriately:
        result = mainPattern(t(1,2), True)
        self.assertEqual(len(result), 1)
        self.assertTrue("c" in result)
        self.assertFalse("a" in result)
        self.assertFalse("b" in result)

    #call with patterns in events
    def test_pattern_call_with_patterns_in_events(self):
        aPattern = time.Pattern(Arc(t(0,1), t(1,2)),
                                [ Event(Arc(t(0,1), t(1,2)), "a"),
                                  Event(Arc(t(1,2),t(1,1)),  "b") ])

        mainPattern = time.Pattern(Arc(t(0,1), t(1,1)),
                                   [ Event(Arc(t(0,1), t(1,1)), "c"),
                                     Event(Arc(t(1,4), t(3,4)),
                                           aPattern, True) ])

        #checks the internal pattern ends appropriately:
        result = mainPattern(t(0,2), True)
        self.assertTrue("c" in result)
        self.assertFalse("a" in result)
        result2 = mainPattern(t(1,4), True)
        self.assertTrue("c" in result2)
        self.assertTrue("a" in result2)
        result3 = mainPattern(t(1,2), True)
        self.assertTrue("c" in result3)
        self.assertFalse("a" in result3)
        self.assertTrue("b" in result3)
        result4 = mainPattern(t(3,4), True)
        self.assertTrue("c" in result4)
        self.assertFalse("a" in result4)
        self.assertFalse("b" in result4)

    #get the key
    def test_pattern_get_key(self):
        aPattern = time.Pattern(Arc(t(3,8),t(6,8)),
                                [ Event(Arc(t(0,1),t(1,1)), "a")])

        self.assertEqual(aPattern.key(), t(3,8))

    #check contains
    def test_pattern_contains(self):
        aPattern = time.Pattern(Arc(t(3,8),t(6,8)),
                                [ Event(Arc(t(0,1),t(1,1)), "a")])

        self.assertTrue(t(4,8) in aPattern)
        self.assertFalse(t(7,8) in aPattern)

    #get the base set
    def test_pattern_denominator_simple(self):
        aPattern = time.parse_string("[ a b c ]")
        denom = aPattern.denominator()
        self.assertEqual(denom, 3)

    def test_pattern_denominator_simple_2(self):
        aPattern = time.parse_string("[a b c d]")
        denom = aPattern.denominator()
        self.assertEqual(denom, 4)

    def test_pattern_denominator_nested(self):
        aPattern = time.parse_string("[a b [c d]]")
        denom = aPattern.denominator()
        self.assertEqual(denom, 6)

    def test_pattern_denominator_parallel(self):
        aPattern = time.parse_string("[a b, c d]")
        denom = aPattern.denominator()
        self.assertEqual(denom, 2)

    #pretty print pattern
    def test_pattern_seq__double_cycle(self):
        p1 = time.parse_string("[a b c d]")
        p2 = time.parse_string("[e f g h]")
        p3 = time.PatternSeq(Arc(t(0,1),t(2,1)),
                             [ p1,p2 ])

        self.assertEqual(p3(t(0,1), True)[0], "a")
        # self.assertEqual(p3(t(1,1), True)[0], "e")
        self.assertEqual(p3(t(7,4), True)[0], "h")

    def test_pattern_iterator(self):
        aPattern = time.parse_string("[a b [c d] e]")
        for i,x in zip([0,1,2,3,4], aPattern.iter()):
            self.assertEqual(x[0], ["a","a",
                                    "b", "b",
                                    "c", "d",
                                    "e", "e"][i])

    def test_pattern_iterator_loop(self):
        aPattern = time.parse_string("[ a b ]")
        for i,x in zip(range(8), aPattern.iter()):
            self.assertEqual(x[0], (["a", "b"] * 4)[i])

    def test_pattern_iterator_events(self):
        aPattern = time.parse_string("[ a b c ]")
        for i,x in zip(range(4), aPattern.iter(False)):
            self.assertIsInstance(x[0], Event)

    def test_pattern_add(self):
        first_pattern = time.parse_string("[a b]")
        second_pattern = time.parse_string("[c d]")
        combined = first_pattern + second_pattern
        for i,x in zip([0, 1, 2, 3, 4, 5],
                       combined.iter()):
            self.assertEqual(x[0],
                            "a b c d a b".split(" ")[i])

    def test_pattern_add_twice(self):
        first_pattern = time.parse_string("[a b]")
        second_pattern = time.parse_string("[c d]")
        combined = first_pattern + second_pattern
        combined_2 = combined + second_pattern
        for i,x in zip([0, 1, 2, 3, 4, 5, 6, 7, 8, 9],
                       combined_2.iter()):
            self.assertEqual(x[0],
                            "a b c d c d a b c d".split(" ")[i])

    def test_pattern_stack(self):
        first_pattern = time.parse_string("[a b]")
        second_pattern = time.parse_string("[c d]")
        combined = first_pattern * second_pattern
        for i,x in zip([0, 1, 2, 3],
                       combined.iter()):
            self.assertEqual(x[0],
                             "a b a b".split(" ")[i])
            self.assertEqual(x[1],
                             "c d c d".split(" ")[i])

    def test_pattern_subtract(self):
        return None

    def test_pattern_format(self):
        return None


    #--------------------
    # PARSER TESTS
    #Parse a pattern
    def test_parse_simple(self):
        aPattern = time.parse_string("[ a b c ]")
        self.assertIsInstance(aPattern, time.Pattern)
        self.assertEqual(len(aPattern.components), 3)

    def test_parse_balance_failure(self):
        with self.assertRaises(Exception):
            time.parse_string("[a b c")

    def test_parse_balance_failure_nested(self):
        with self.assertRaises(Exception):
            time.parse_string("[a b [c d]")

    #Parse a nested pattern
    def test_parse_nested_simple(self):
        aPattern = time.parse_string("[ a b [c d]]")
        self.assertEqual(len(aPattern.components), 3)
        self.assertIsInstance(aPattern.components[2].values, time.Pattern)

    def test_parse_parallel_nested(self):
        aPattern = time.parse_string("[ a b , [c d] e]")
        self.assertEqual(len(aPattern.components), 2)
        self.assertEqual(len(aPattern.components[1].values.components), 2)

    #parse a pretty printed pattern

if __name__ == "__main__":
      #use python $filename to use this logging setup
      LOGLEVEL = logging.INFO
      logFileName = "log.test_time"
      logging.basicConfig(filename=logFileName, level=LOGLEVEL, filemode='w')
      console = logging.StreamHandler()
      console.setLevel(logging.INFO)
      logging.getLogger().addHandler(console)
      unittest.main()
      #reminder: user logging.getLogger().setLevel(logging.NOTSET) for log control



#!/usr/bin/env python3
"""
Doot : An Opinionated refactor of the Doit Task Runner.

"""
##-- std imports
from __future__ import annotations

import logging as logmod
import pathlib as pl
from typing import Final, Any, assert_type
from importlib.resources import files
import sys
##-- end std imports

##-- imports
import tomlguard as TG
##-- end imports

##-- data
data_path      = files("doot.__data")
constants_file = data_path.joinpath("constants.toml")
aliases_file   = data_path.joinpath("aliases.toml")
##-- end data

##-- logging
logging         = logmod.getLogger(__name__)
printer         = logmod.getLogger("doot._printer")
##-- end logging

# Global, single points of truth:
__version__          : Final[str]         = "0.6.0"

CONSTANT_PREFIX      : Final[str]         = "doot.constants"
ALIAS_PREFIX         : Final[str]         = "doot.aliases"
TOOL_PREFIX          : Final[str]         = "tool.doot"

config               : TG.TomlGuard       = TG.TomlGuard() # doot config
constants            : TG.TomlGuard       = TG.TomlGuard.load(constants_file).remove_prefix(CONSTANT_PREFIX)
aliases              : TG.TomlGuard       = TG.TomlGuard()
locs                 : DootLocData        = None # DootLocations(pl.Path()) # registered locations
args                 : TG.TomlGuard       = TG.TomlGuard() # parsed arg access
report               : Reporter_i         = None

_configs_loaded_from : list[pl.Path]      = []

def setup(targets:list[pl.Path]|None=None, prefix:str|None=TOOL_PREFIX) -> tuple[TG.TomlGuard, DootLocData]:
    """
      The core requirement to call before any other doot code is run.
      loads the config files, so everything else can retrieve values when imported.

      `prefix` removes a prefix from the loaded data.
      eg: 'tool.doot' for if putting doot settings in a pyproject.toml
    """
    global config, _configs_loaded_from
    targets : list[pl.Path] = [pl.Path(x) for x in targets or constants.paths.DEFAULT_LOAD_TARGETS]
    logging.debug("Loading Doot Config, version: %s targets: %s", __version__, targets)
    if bool(config):
        printer.warning("doot.setup called even though doot is already set up")

    if bool(targets) and not all([isinstance(x, pl.Path) for x in targets]):
        raise TypeError("Doot Config Targets should be pathlib.Path's", targets)

    if not any([x.exists() for x in targets]):
        raise doot.errors.DootMissingConfigError("No Doot data found")

    existing_targets       = [x for x in targets if x.exists()]

    try:
        config = TG.load(*existing_targets)
    except OSError:
        logging.error("Failed to Load Config Files: %s", existing_targets)
        raise doot.errors.DootError()

    config = config.remove_prefix(prefix)
    _load_locations()
    _load_constants()
    _load_aliases()
    _update_import_path()

    _configs_loaded_from   = existing_targets

    return config, locs

def _load_constants():
    """ Load the override constants if the loaded base config specifies one"""
    global constants
    update_file = config.on_fail(None).settings.general.constants_file()
    if update_file:
        constants = TG.TomlGuard.load(pl.Path(update_file)).remove_prefix(CONSTANT_PREFIX)

def _load_aliases():
    """ Load plugin aliases """
    global aliases
    update_file = config.on_fail(aliases_file).settings.general.aliases_file()
    data        = TG.TomlGuard.load(pl.Path(update_file)).remove_prefix(ALIAS_PREFIX)
    # Flatten the lists
    flat = {}
    for key,val in data:
        flat[key] = {k:v for x in val for k,v in x.items()}

    # Then override with config specified:
    for key,val in config.plugins:
        flat[key].update(dict(val))

    aliases = TG.TomlGuard(flat)

def _load_locations():
    """ Load and update the DootLocations db """
    global locs
    from jgdv.locations.locations import JGDVLocations
    locs   = JGDVLocations(pl.Path.cwd())
    # Load Initial locations
    for loc in config.on_fail([]).locations():
        locs.update(loc)

def _update_import_path():
    """ Add locations to the python path for task local code importing  """
    task_sources = config.on_fail([locs[".tasks"]], list).settings.tasks.sources(wrapper=lambda x: [locs[y] for y in x])
    task_code    = config.on_fail([locs[".tasks"]], list).settings.tasks.code(wrapper=lambda x: [locs[y] for y in x])
    for source in set(task_sources + task_code):
        if source.exists() and source.is_dir():
            logging.debug("Adding task code directory to Import Path: %s", source)
            sys.path.append(str(source))



#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import pyparsing as pp

s         = pp.Suppress
op        = pp.Optional
orm       = pp.OneOrMore
zrm       = pp.ZeroOrMore

ln                = s(pp.White("\n\r", max=1).set_whitespace_chars("\t "))
manyLine          = s(pp.White("\n\r", min=1).set_whitespace_chars("\t "))
emptyLine         = s(ln + manyLine)
opLn              = op(ln)
tab               = pp.White(TAB_S, min=2).set_whitespace_chars("\r\n")

emptyLine.set_name("emptyLine")
ln.set_name("line")
opLn.set_name("OptionalLine")
tab.set_name("tab")

def gap_fail_action(s, loc, expr, err):
    logging.warning("{}\n{}".format(str(err), err.markInputline()))



"""

"""
##-- imports
from __future__ import annotations

import importlib
import logging as logmod
import re
import traceback
from datetime import datetime
from enum import Enum
from os.path import abspath, exists, expanduser, split, splitext

##-- end imports

logging = logmod.getLogger(__name__)
config  = acab.config

@register
def do_init(self, line):
    """
    Specify the Engine to initialise.
    Imports the module, and uses the final component as the Engine Class.
    eg: acab.modules.engines.trie_engine.TrieEngine -> TrieEngine

    A Question mark at the end of the line signals to inspect the module
    for potential constructors:
    eg: acab.modules.engines.configured?
    """
    if not bool(line.strip()):
        line = self.state.engine_str

    # if:  "init line?", then return applicable functions
    if line[-1] == "?":
        return init_inspect(line[:-1])

    logging.info("Initialising Engine: {}".format(line))

    try:
        mod_str = splitext(line)[0]
        mod = importlib.import_module(mod_str)
        # TODO ask for confirmation?
        # Note: not init_module.{} because of split*ext*
        # build engine. needs to be a 0 arg constructor
        spec        = getattr(mod, line.split(".")[-1])
        is_type     = isinstance(spec, type)
        is_sub      = is_type and issubclass(spec, AcabEngine_i)
        is_callable = callable(spec)
        if (not is_type) and isinstance(spec, AcabEngine_i):
            self.state.engine = spec
        elif (is_type and is_sub) or callable(spec):
            self.state.engine = spec()
        else:
            raise AcabConfigException(f"Unknown Engine Spec Form: {spec}")

        # TODO add bad words from repl:
        # self.state.engine.parser.set_word_exclusions(self.completenames(""))

        self.state.ctxs = None
        logging.info("Engine Initialisation Complete")
    except Exception as err:
        logging.error(f"Failed to initialise engine: {line}", exc_info=err)



#!/usr/bin/env python3
"""


See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import importlib

def init_inspect(mod_str):
    """
    Import and Inspect the passed in module for potential constructor functions
    to init with
    """
    mod = importlib.import_module(mod_str)
    try:
        not_dunders    = [getattr(mod, x) for x in dir(mod) if "__" not in x]
        not_modules    = [x for x in not_dunders if not isinstance(x, ModuleType)]
        correct_module = [x for x in not_modules if mod_str in x.__module__]
        funcs = [x for x in correct_module if isinstance(x, FunctionType)]
        engines        = [x for x in correct_module if isinstance(x, type) and issubclass(x, AcabEngine_i)]
        total = funcs + engines

        if not bool(total):
            return print(f"No Available Constructors in {mod_str}")

        print(f"Potential Constructors in {mod_str}:")
        for x in total:
            print(f"-- {x.__name__}")
    except:
        breakpoint()



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

import importlib
from tomlguard import TomlGuard

PAD           : Final[int]  = 15
PARAM_ASSIGN_PREFIX         = "--" # doot.constants.patterns.PARAM_ASSIGN_PREFIX:
parse_err : type(Exception) = Exception # doot.errors.DootParseError

@dataclass
class JGDVParamSpec:
    """ Describes a command line parameter to use in the parser
      When `positional`, will not match against a string starting with `prefix`
      consumed in doot._abstract.parser.ArgParser_i's
      produced using doot._abstract.parser.ParamSpecMaker_m classes,
      like tasks, and jobs
    """
    name              : str                       = field()
    type              : type                      = field(default=bool)

    prefix            : str                       = field(default="-")

    default           : Any                       = field(default=False)
    desc              : str                       = field(default="An undescribed parameter")
    constraints       : list                      = field(default_factory=list)
    invisible         : bool                      = field(default=False)
    positional        : bool|int                  = field(default=False)
    _short            : None|str                  = field(default=None)

    _repeatable_types : ClassVar[list[Any]]       = [list, int]

    separator         : str                       = field(default="=")
    _consumed         : int                       = field(default=0)

    @classmethod
    def build(cls, data:TomlGuard|dict) -> JGDVParamSpec:
        param =  cls(**data)
        match param:
            case JGDVParamSpec(type="int"):
                param.type = int
            case JGDVParamSpec(type="float"):
                param.type = float
            case JGDVParamSpec(type="bool"):
                param.type = bool
            case JGDVParamSpec(type="str"):
                param.type = str
            case JGDVParamSpec(type="list"):
                param.type = list

        if param.default == "None":
            param.default = None

        return param

    @staticmethod
    def key_func(x):
        return (x.positional != 0, x.prefix)

    @property
    def short(self):
        if self.positional:
            return self.name

        if self._short:
            return self._short

        return self.name[0]

    @property
    def inverse(self):
        return f"no-{self.name}"

    @property
    def repeatable(self):
        return self.type in JGDVParamSpec._repeatable_types and not self.positional

    @property
    def key_str(self):
        if self.invisible or self.positional:
            return ""

        if self.prefix == PARAM_ASSIGN_PREFIX:
            return f"{self.prefix}{self.name}{self.separator}"

        return f"{self.prefix}{self.name}"

    @property
    def short_key_str(self):
        if self.invisible or self.positional:
            return ""

        if self.prefix == PARAM_ASSIGN_PREFIX:
            return f"{self.prefix}{self.name[0]}{self.separator}"

        return f"{self.prefix}{self.name[0]}"

    def _split_name_from_value(self, val):
        match self.positional:
            case False:
                return val.removeprefix(self.prefix).split(self.separator)
            case True if isinstance(val, list):
                return (self.name, val)
            case True | int():
                return (self.name, [val])

    def __eq__(self, val) -> bool:
        """ test to see if a cli argument matches this param """
        if 0 < self.positional <= self._consumed:
            return False

        match val, self.positional:
            case JGDVParamSpec(), _:
                return val is self
            case str(), False:
                [head, *_] = self._split_name_from_value(val)
                return head in [self.name, self.short, self.inverse]
            case str(), True:
                return not val.startswith(self.prefix)
            case str(), int():
                return not val.startswith(self.prefix)
            case _, _:
                return False

    def __str__(self):
        if self.invisible:
            return ""

        parts = [self.key_str or f"[{self.name}]"]

        parts.append(" " * (PAD-len(parts[0])))
        match self.type:
            case type() if self.type == bool:
                parts.append(f"{'(bool)': <10}:")
            case str() if bool(self.default):
                parts.append(f"{'(str)': <10}:")
            case str():
                parts.append(f"{'(str)': <10}:")
            case _:
                pass

        parts.append(f"{self.desc:<30}")
        pad = " "*max(0, (85 - (len(parts)+sum(map(len, parts)))))
        match self.default:
            case None:
                pass
            case str():
                parts.append(f'{pad}: Defaults to: "{self.default}"')
            case _:
                parts.append(f"{pad}: Defaults to: {self.default}")

        if self.constraints:
            parts.append(": Constrained to: {self.constraints}")
        return " ".join(parts)

    def __repr__(self):
        if self.positional:
            return f"<ParamSpec: {self.name}>"
        return f"<ParamSpec: {self.prefix}{self.name}>"

    def _calc_positional_consumption(self, focus, args):
        pop_count     = 1
        prefixed      = focus.startswith(self.prefix) # form of -param
        is_assign     = self.separator in focus       # form of --param=arg
        key, vals     = None, []

        # Figure out the key and value
        match prefixed, is_assign:
            case _, True if self.prefix != PARAM_ASSIGN_PREFIX:
                raise parse_err("Assignment parameters should be prefixed with the PARAM_ASSIGN_PREFIX", doot.constants.patterns.PARAM_ASSIGN_PREFIX)
            case True, False if self.type.__name__ != "bool" and not bool(args):
                raise parse_err("key lacks a following value", focus, self.type.__name__)
            case True, True: # --key=val
                key, val = focus.split(self.separator)
                key = key.removeprefix(self.prefix)
                vals.append(val)
            case True, False if self.type.__name__ == "bool": # --key
                key = focus.removeprefix(self.prefix)
            case True, False: # -key val
                key = focus.removeprefix(self.prefix)
                vals.append(args[1])
                pop_count = 2

        return key, vals, pop_count

    def _add_non_positional_value(self, data:dict, *, key:str=None, vals:list[str]=None) -> bool:
        """ if the given value is suitable, add it into the given data
        takes separated key, values,
        and the key has had the prefix stripped
        """
        vals = vals or []
        # TODO if constraints, check against them
        logging.debug("Matching: %s : %s : %s", self.type.__name__, key, vals)

        # Use type.__name__ because you can't match on type. ("case str" fails, expecting "case str()")
        match self.type.__name__:
            case "list":
                if self.name not in data or data[self.name] == self.default:
                    data[self.name] = []
                data[self.name] += vals
            case "set":
                if self.name not in data or data[self.name] == self.default:
                    data[self.name] = set()
                data[self.name].update(vals)
            case _ if data.get(self.name, self.default) != self.default:
                raise parse_err("Trying to re-set an arg already set: %s : %s", self.name, vals)
            ##-- handle bools and inversion
            case "bool" if bool(vals):
                raise parse_err("Bool Arguments shouldn't have values: %s : %s", self.name, vals)
            case "bool" if key == self.inverse:
                data[self.name] = False
            case "bool":
                data[self.name] = True
            ##-- end handle bools and inversion
            case _ if not bool(vals):
                raise parse_err("Non-Bool Arguments should have values: %s : %s", self.name, vals)
            case _ if len(vals) == 1:
                data[self.name] = self.type(vals[0])
            case _:
                raise parse_err("Can't understand value: %s : %s", self.name, vals)

        return data[self.name] != self.default

    def _add_positional_value(self, data, *, key:str=None, vals:list[str]=None) -> int:
        pop_count = 1
        end_index = None if "--" not in vals else vals.index("--")

        match self.positional:
            case True if self.type != list:
                data[self.name] = vals[0]
            case True:
                consume_these      = vals[:end_index]
                pop_count          = len(consume_these)
                data[self.name]    = consume_these
            case 1:
                data[self.name] = self.type(vals[0])
                pop_count = 1
            case x:
                t = min(x, end_index or len(vals))
                consume_these      = vals[:t]
                pop_count          = len(consume_these)
                data[self.name]    = consume_these

        if 1 < pop_count < self.positional:
            raise parse_err("Not Enough positional args provided", self.name, self.positional, vals)
        elif 1 < self.positional < pop_count:
            raise parse_err("Too Many positional args provided", self.name, self.positional, vals)
        elif 1 < pop_count and self.type != list:
            raise parse_err("Multi positional args should be of type list", self.name, self.positional, self.type)

        return pop_count

    def maybe_consume(self, args:list[str], data:dict) -> int:
        """
          Given a list of args, possibly add a value to the data.
          operates *in place* on both the args list and the data.
          return True if _consumed a value

          handles:
          ["--arg=val"],
          ["-arg", "val"],
          ["val"],     (if positional=True)
          ["-arg"],    (if type=bool)
          ["-no-arg"], (if type=bool)
        """
        if not bool(args) or data is None or args[0] != self:
            return 0
        if 0 < self._consumed:
            return self._consumed

        pop_count     = 0
        focus         = args[0]
        is_positional = bool(self.positional)

        if is_positional:
            pop_count = self._add_positional_value(data, key=self.name, vals=args)
        else:
            key, vals, pop_count = self._calc_positional_consumption(focus, args)
            self._add_non_positional_value(data, key=key, vals=vals)

        # data has been added, so remove it from the input list
        logging.debug("Arg: %s consuming count %s", self.name, pop_count)
        for x in range(pop_count):
            args.pop(0)

        self._consumed += pop_count
        return self._consumed



#!/usr/bin/env python3
"""

See EOF for license/metadata/notes as applicable
"""

##-- builtin imports
from __future__ import annotations

# import abc
import datetime
import enum
import functools as ftz
import itertools as itz
import logging as logmod
import pathlib as pl
import re
import time
import types
import weakref
# from copy import deepcopy
# from dataclasses import InitVar, dataclass, field
from typing import (TYPE_CHECKING, Any, Callable, ClassVar, Final, Generic,
                    Iterable, Iterator, Mapping, Match, MutableMapping,
                    Protocol, Sequence, Tuple, TypeAlias, TypeGuard, TypeVar,
                    cast, final, overload, runtime_checkable, Generator)
from uuid import UUID, uuid1

##-- end builtin imports

##-- lib imports
import more_itertools as mitz
##-- end lib imports

##-- logging
logging = logmod.getLogger(__name__)
##-- end logging

from random import random
import struct

logging = logmod.getLogger(__name__)

# BigEnd, continue?, msglen
head = struct.Struct(">?L")

kg =  b'keep going'
fin = b'finished'

served = 0

async def still_alive():
    while True:
        print("Still Alive")
        await asyncio.sleep(1)

async def handle_connection(reader, writer):
    message = None
    addr    = writer.get_extra_info('peername')
    print("Connection Opened: ", addr)

    count = 0
    while True:
        data    = await reader.read(head.size)
        match head.unpack(data):
            case True, size:
                count += 1
                print("Reading: ", size)
                data    = await reader.read(size)
                print("Message: ", data.decode())

                if count > 3:
                    response = head.pack(False, len(fin))
                    writer.write(response)
                    writer.write(kg)
                else:
                    response = head.pack(True, len(kg))
                    writer.write(response)
                    writer.write(kg)
                await writer.drain()
            case False, 0:
                break
            case _:
                raise Exception("Bad message received")

    print("Close the connection")
    writer.close()
    global served
    served += 1

async def main():
    server = await asyncio.start_server(handle_connection, '127.0.0.1', 8888)
    addr   = server.sockets[0].getsockname()
    print(f'Serving on {addr}')

    sa_task = asyncio.create_task(still_alive())
    async with server:
        await server.serve_forever()

if __name__ == '__main__':
    asyncio.run(main())



#!/usr/bin/env python3
##-- imports
from __future__ import annotations

import pathlib as pl
import datetime
import json
import logging as logmod
from collections import defaultdict
from dataclasses import InitVar, dataclass, field
from typing import (Any, Callable, ClassVar, Dict, Generic, Iterable, Iterator,
                    List, Mapping, Match, MutableMapping, Optional, Sequence,
                    Set, Tuple, TypeVar, Union, cast)
from uuid import uuid1

import networkx as nx
##-- end imports

class OrgBuilderBase:

    link_pattern        : str = "[[{}]]"
    named_link_pattern  : str = "[[{}][{}]]"
    named_file_pattern  : str = "[[file:{}][{}]]"




#!/usr/bin/env python
"""
Tagset Reading

"""
##-- imports
from __future__ import annotations

import logging as logmod
import re
from dataclasses import dataclass, field
from typing import (Any, Callable, ClassVar, Dict, Generic, Iterable, Iterator,
                    List, Mapping, Match, MutableMapping, Optional, Sequence,
                    Set, Tuple, TypeVar, Union, cast, Final, TypeAlias)
import re
##-- end imports

import networkx as nx

logging = logmod.getLogger(__name__)

IGNORE_REPLACEMENTS = ["TO_CHECK"]

TAG_NORM    : Final[re.Pattern] = regex.compile(" +")
ORG_PATTERN : Final[str]        = r"^\*\*\s+.+?\s+:(\S+):$"
ORG_SEP     : Final[str]        = ":"

Tag : TypeAlias = str

class TagGraph:

    def __init__(self):
        self.graph : nx.Graph = nx.Graph()

    def extract_bibtex(self, db:bibtexparser.BibtexDatabase) -> TagFile:
        logging.info("Processing Bibtex: %s", len(db.entries))

        proportion = int(len(db.entries) / 10)
        count = 0
        total = TagFile()

        for i, entry in enumerate(db.entries):
            if i % proportion == 0:
                logging.info("%s/10 Complete", count)
                count += 1

            #get tags
            total.update(self.link(entry['tags']))

        return total

    def extract_org(self, org_files:List[pl.Path], tag_regex=None) -> TagFile:
        logging.info("Extracting data from orgs")
        tag_regex = tag_regex or ORG_PATTERN

        ORG_TAG_REGEX = regex.compile(tag_regex)
        total = TagFile()

        for org in org_files:
            #read
            text = []
            with open(org,'r') as f:
                text = f.readlines()

            #line by line
            for line in text:
                tags = ORG_TAG_REGEX.findall(line)
                e_tags = []
                if not bool(tags):
                    continue

                e_tags = [x for x in tags[0].split(ORG_SEP)]
                total.update(self.link(e_tags))

        return total

    def extract_bookmark(self, bkmk_files: List[pl.Path]) -> TagFile:
        total = TagFile()
        for bkmk_f in bkmk_files:
            bkmks = BookmarkCollection.read(bkmk_f)

            for bkmk in bkmks:
                tags          = bkmk.tags
                total.update(self.link(tags))

        return total

    def link(self, tags:Iterable[Tag]) -> Iterable[Tag]:
        """
        Add a set of tags to the graph, after normalising
        """
        norm_tags = [TAG_NORM.sub("_", x.strip()) for x in tags if bool(x)]
        remaining = norm_tags[:]

        [self.graph.add_node(x, count=0) for x in norm_tags if x not in self.graph]

        for tag in norm_tags:
            self.graph.nodes[tag]['count'] += 1
            remaining.remove(tag)
            edges_to_increment = [(tag, y) for y in remaining]
            for u,v in edges_to_increment:
                if not self.graph.has_edge(u,v):
                    self.graph.add_edge(u,v, weight=0)

                self.graph[u][v]['weight'] += 1

        return norm_tags

    def write(self, target:pl.Path):
        nx.write_weighted_edgelist(self.graph, str(target))

    def __str__(self):
        keys    = self.tags
        tag_str = "\n".join(["{} : {}".format(k, self.graph.nodes[k]['count']) for k in keys])
        return tag_str

    @property
    def tags(self) -> TagFile:
        result = TagFile()
        for tag in self.graph.nodes:
            result.set_count(tag, self.graph.nodes[tag]['count'])
        return result

    def get_count(self, tag:Tag):
        if tag not in self.graph:
            return 0
        return self.graph[tag]['count']
